#ifndef LINKFILE_H
#define LINKFILE_H

/*
C version of various pedigree and locus handling functions, swiped
for DOLINK so that locus input/output routines can be used
*/

#if defined DCERRORHPP && ! defined USEDCERROR
#define USEDCERROR
/*
use dcerror class for error reporting
define USEDCERROR if dcerror.hpp has already been included, but
alternatively USEDCERROR can be defined as a compiler switch

if we're doing this we don't want the macro below to be seen by
any C++ classes which might have their own error() functions.
*/
#endif

#ifndef USEDCERROR
#define error(s,t) error_func(__LINE__,__FILE__,s,t)
#endif

#define MAXALL_LF 30
#define MAXFACT_LF 30
#define MAXLOCI_LF 20
#define MAXIND_LF 100
#define MAXLIAB_LF 20
#define MAXCHILD_LF 20
#define MAXSPOUSE_LF 20
#define MAXNAME_LF 20
#ifdef __STDC__
#define CAN_PROTOTYPE 1
#endif
#if __cplusplus
#define CAN_PROTOTYPE 1
extern "C" {
#endif

#ifndef CAN_PROTOTYPE
#define CAN_PROTOTYPE 0
#endif

#if CAN_PROTOTYPE
#define PROTO(args) args
#else
#define PROTO(args) ()
#endif

enum { DCL_AFF=1,DCL_BINARY,DCL_NUMBERED,DCL_QUAN } ;

struct dc_locus_t {
int type;
union {
int aff[2];
int all[2];
char bin[MAXFACT_LF];
float qval;
} data;
};

typedef struct dc_locus_t dc_locus;

struct aff_def_t {
int nliab;
float pen[MAXLIAB_LF][3];
float male_pen[MAXLIAB_LF][2];
};

typedef struct aff_def_t aff_def;

struct bin_def_t {
int nfact;
char fact[MAXALL_LF][MAXFACT_LF];
};

typedef struct bin_def_t bin_def;

struct quan_def_t {
float mAA,maa,dom,var,hetmult;
};

typedef struct quan_def_t quan_def;

struct dc_locus_def_t {
int type;
float freq[MAXALL_LF];
int nalls;
int isrisk,riskall;
union {
aff_def aff;
bin_def bin;
quan_def quan;
} info;
};

typedef struct dc_locus_def_t dc_locus_def;

struct dc_person_t;

struct dc_mating_t {
struct dc_person_t *father,*mother,*child[MAXCHILD_LF];
int nchild;
};

typedef struct dc_mating_t dc_mating;

struct dc_person_t {
char fname[MAXNAME_LF+1];
int idnum,fidnum,midnum,isproband,sex;
int nspouse;
dc_mating *parents,*spouse[MAXSPOUSE_LF];
dc_locus locus[MAXLOCI_LF];
};

typedef struct dc_person_t dc_person;

extern error_func PROTO((unsigned l,char *f,char *s1,char *s2));
extern locus_def_input PROTO((FILE *fp,dc_locus_def *def,int sexlinked,int risklocus));
extern locus_def_output PROTO((FILE *fp,dc_locus_def *def,int sexlinked));
extern read_person PROTO((FILE *fp,dc_person *ind,dc_locus_def *def,int nloci));
extern read_pedigree PROTO((FILE *fp,dc_person *ped,dc_locus_def *def,int nloci));
extern set_parents PROTO((dc_person *child,dc_person *ped,int count));
extern free_matings PROTO((dc_person *ped,int count));
extern make_matings PROTO((dc_person *ped,int count));
extern locus_input PROTO((dc_locus *locus,dc_locus_def *def,char **line));
extern read_data_file PROTO((FILE *fp,dc_locus_def *def,int *sexlinked,int *risklocus));



#if __cplusplus
}
#endif

#endif
