/* PEDIGREE.CPP */
/* Copyright Dave Curtis 1991 */
#include "pedigree.hpp"
#include <stdlib.h>
#include <string.h>
#include "dcerror.hpp"
#ifdef MDB
#include <mem.h>
#endif


person *person::find_proband()
{
if (!parents) return this;
else if (!get_mother()->parents) return get_father()->find_proband();
else return get_mother()->find_proband();
}

person *pedigree::get_anc(int i)
{ member(i)->clear(); return member(i); }

void pedigree::clear_pointers()
{
if (!pointers_done) return;
else pointers_done=0;
for (int i=0;i<count;++i)
  member(i)->clear_pointers();
}

void person::clear_pointers()
{
int j;
parents=0;
if (gender=='M')
  for (j=0;j<numspo;++j)
      {
      if (--spouse[j]->numrefs==0)
        delete spouse[j];
      spouse[j]=0;
      }
numch=numspo=0;
}

person::person()
{
//id=fid=mid=0;
diag=numch=numspo=0;
parents=0;
gender='X';
int i;
for (i=0;i<MAXSPOUSE;++i)
  spouse[i]=0;
}

person::~person()
{
if (gender=='M')
  for (int i=0;i<numspo; ++i)
    if (--spouse[i]->numrefs==0) delete spouse[i];
}

person& person::operator=(const person &old)
{
diag=old.diag;
numch=old.numch;
numspo=old.numspo;
parents=old.parents;
gender=old.gender;
int i;
for (i=0;i<MAXSPOUSE;++i)
  if (spouse[i])
    {
    if (gender=='M' && --spouse[i]->numrefs==0) delete spouse[i];
    spouse[i]=0;
    }
for (i=0;i<numspo;++i)
   { spouse[i]=old.spouse[i]; if (gender=='M') ++spouse[i]->numrefs; }
if (!id.assign(old.id) || !mid.assign(old.mid) || !fid.assign(old.fid))
   dcerror(ENOMEM,"Out of memory in person::operator=()");
return *this;
}

pedigree::do_pointers()
{
if (pointers_done) return 1;
validate();
int i,j;
person *father,*mother,*mi,*temp;
for (i=0;i<count && valid;++i)
  {
  mi=member(i);
  father=mother=0;
//       if (mi->fid.is_null() || mi->mid.is_null()) continue;  // no parent
  if (!mi->fid.is_null())
    {
    if ((j=which(mi->fid))!=-1)
      {
      father=member(j);
      if (father->gender!='M')
        { dcerror(1,"\n%s%s\n",mi->id," has a non-male father"); invalidate(); }
      }
    else
      {
      dcerror(2,"Adding %s, father of %s\n",mi->fid,mi->id);
      if ((temp=(person*)new_own_type())==NULL ||
            !temp->id.assign(mi->fid) ||
            !add_member(temp))
              { dcerror(ENOMEM,"Out of memory in pedigree::do_pointers()"); invalidate(); }
      else
        {
        father=member(count-1);
        father->diag=DEAD;
        father->gender='M';
        }
      }
    }
  if (!mi->mid.is_null())
    {
    if ((j=which(mi->mid))!=-1)
      {
      mother=member(j);
      if (mother->gender!='F')
        { dcerror(1,"\n%s%s\n",mi->id," has a male mother"); invalidate(); }
      }
    else
      {
      dcerror(2,"Adding %s, mother of %s\n",mi->mid,mi->id);
      if ((temp=(person*)new_own_type())==NULL ||
            !temp->id.assign(mi->mid) ||
            !add_member(temp))
              { dcerror(ENOMEM,"Out of memory in pedigree::do_pointers()"); invalidate(); }
      else
        {
        mother=member(count-1);
        mother->diag=DEAD;
        mother->gender='F';
        }
      }
    }
  if (!father != !mother)
          { dcerror(2,"%s has only one parent",(char*)mi->id); invalidate(); }
  if (!father || !mother) continue;
  for (j=0;j<father->numspo;++j)
    if (father->get_spouse(j)==mother)
      goto got_mating;
  if (j>=MAXSPOUSE)
    { dcerror(3,"%s has more than maximum permitted number of spouses (%d)",
       (char*)father->id,MAXSPOUSE); invalidate(); continue; }
  if ((father->spouse[j]=new mating(father,mother))==0)
    {
    dcerror(ENOMEM,"Out of memory in pedigree::do_pointers()");
    invalidate();
    continue;
    }
  ++father->numspo;
  if (mother->numspo>=MAXSPOUSE)
    { dcerror(3,"%s has more than maximum permitted number of spouses (%d)",
       (char*)mother->id,MAXSPOUSE); invalidate(); continue; }
  mother->spouse[mother->numspo++]=father->spouse[j];
  got_mating:
     if (father->spouse[j]->get_numch()>=MAXCHILD)
    { dcerror(3,"Mating between %s and %s has more than maximum permitted number of children (%d)",
       (char*)father->id,(char*)mother->id,MAXCHILD); invalidate(); continue; }
     father->spouse[j]->add_child(mi);
     mi->parents=father->spouse[j];
  }
free_hash();
if (is_valid()) pointers_done=1;
return !dcerror.stat();
}




