/* PEDIGREE.HPP */
/* Copyright Dave Curtis 1991 */
#ifndef PEDIGREEHPP
#define PEDIGREEHPP 1

#include <hasharry.hpp>
#include <dcstr.hpp>

class person;
class mating;
class locus;
class linksys_person;
class tab_person;
class ped_person;
class ppd_person;
class table;
class table_item;
class pedigree;
class tree_data_file;
class sdf_data_file;
class linksys_data_file;
class editable_person;

#define DEAD 100
#ifdef DOS386
#define MAXCHILD 40
#define MAXSPOUSE 40
#else
#define MAXCHILD 20
#define MAXSPOUSE 20
#endif

class mating {
friend class person;
person *father,*mother;
person *child[MAXCHILD];
int numch;
int numrefs;
int status;
public:
get_stat() const { return status; }
void set_stat(int s) { status=s; }
get_numch() const { return numch; }
person *get_child(int ch) const { return child[ch]; }
mating(person *f, person *m)
  { father=f; mother=m; numch=0; numrefs=1; status=0; }
void add_child(person *p) { child[numch++]=p; }
};

class person :public data_object
{
friend class pedigree;
friend class linksys_pedigree;
friend class tab_pedigree;
friend class ped_person;
friend class ppd_person;
protected:
dcstring id,fid,mid;
char gender;
mating *parents;
mating *spouse[MAXSPOUSE];
int numch,numspo,diag;
void clear_pointers();
public:
char *label() { return id; }
person *find_proband();
person();
person &operator=(const person &old);
virtual ~person(); // so tab_person is destroyed
print(char *s,int line,int length=-1);
char *get_id() const { return id; }
char *get_father_id() const { return fid; }
char *get_mother_id() const { return mid; }
char get_gender() const { return gender; }
mating *get_mating(int i) const { return spouse[i]; }
person *get_spouse(int i) const
  { return spouse[i]?gender=='M'?spouse[i]->mother:spouse[i]->father:0; }
// person *get_child(int i) { return child[i]; }
person *get_father() const { return parents?parents->father:0; }
person *get_mother() const { return parents?parents->mother:0; }
int get_diag() { return diag; }
int get_numch() { return numch; }
int get_numspo() { return numspo; }
virtual void clear() { ; }
//virtual void clear()=0;
virtual void *base() { return this; }
//virtual operator void *() { return this; }
};

declare (ghash_dcarray,person);

class pedigree : public ghash_dcarray(person)
{
protected:
char valid;
char pointers_done;
void validate() { valid=1; }
void invalidate() { valid=0; }
public:
int is_valid() const { return valid; }
are_pointers_done() const { return pointers_done; }
pedigree(int dim=20) : ghash_dcarray(person)(dim) { pointers_done=0; }
~pedigree() {;}
virtual do_pointers();
void clear_pointers();
person *get_anc(char *name);
person *get_anc(int i);
};

#endif

