/* PEDLINK.CPP */
/* Copyright Dave Curtis 1991 */
#include "pedlink.hpp"
#include "dcwindow.hpp"
#include "dcerror.hpp"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "hashfile.hpp"
#include "dcutils.hpp"
#ifdef MDB
#include <mem.h>
#endif

#ifndef max
#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))
#endif

#define hereok() printf("Got to line %d OK\n",__LINE__)


extern "C" {
int locus_compare(locus **l1,locus **l2)
{
if ((*l1)->type!=(*l2)->type)
  return (*l1)->type<(*l2)->type?-1:1;
else if ((*l1)->chromosome!=(*l2)->chromosome)
  {
  if ((*l1)->chromosome==-1) return 1;
  else if ((*l2)->chromosome==-1) return -1;
  else return (*l1)->chromosome>(*l2)->chromosome?1:-1;
  }
else if (!(*l1)->chromosome)
  return (strcmp((*l1)->name,(*l2)->name));
else
  {
  char *ptr1,*ptr2;
  if ((*l1)->chromosome==-1)
    {
    ptr1=strpbrk((*l1)->comment,"Xx")+1;
    ptr2=strpbrk((*l2)->comment,"Xx")+1;
    }
  else
   {
   ptr1=strpbrk((*l1)->comment,"Cc");
   while (isdigit(* ++ptr1)) ;
   ptr2=strpbrk((*l2)->comment,"Cc");
   while (isdigit(* ++ptr2)) ;
   }
 if (toupper(*ptr1)=='Q' && toupper(*ptr2)== 'Q')
   {
   if (atoi(ptr1+1)==atoi(ptr2+1))
     {
     if ((strncmp(ptr1+1,"ter",3)&&strncmp(ptr1+1,"TER",3)) !=
        (strncmp(ptr2+1,"ter",3)&&strncmp(ptr2+1,"TER",3)))
        return (strncmp(ptr1+1,"ter",3)&&strncmp(ptr1+1,"TER",3))?1:-1;
     else
        return (strcmp((*l1)->name,(*l2)->name));
     }
   else if (!atoi(ptr1+1)) return 1;
   else if (!atoi(ptr2+1)) return -1;
   else return atoi(ptr1+1)>atoi(ptr2+1)?1:-1;
   }
 else if (toupper(*ptr1)=='P' && toupper(*ptr2)== 'P')
   {
   if (!(strncmp(ptr1+1,"ter",3)&&strncmp(ptr1+1,"TER",3)) ||
        !(strncmp(ptr2+1,"ter",3)&&strncmp(ptr2+1,"TER",3)))
        return
          (!(strncmp(ptr1+1,"ter",3)&&strncmp(ptr1+1,"TER",3)) &&
            !(strncmp(ptr2+1,"ter",3)&&strncmp(ptr2+1,"TER",3)))?
               strcmp((*l1)->name,(*l2)->name)
               :!(strncmp(ptr1+1,"ter",3)&&strncmp(ptr1+1,"TER",3))?-1:1;
   else if (atoi(ptr1+1)==atoi(ptr2+1))
       return strcmp((*l1)->name,(*l2)->name);
   else return atoi(ptr1+1)<atoi(ptr2+1)?1:-1;
   }
 else if (toupper(*ptr1)=='P') return -1;
 else if (toupper(*ptr2)=='P') return 1;
 else if (toupper(*ptr1)=='Q') return -1;
 else if (toupper(*ptr2)=='Q') return 1;
 else return strcmp((*l1)->name,(*l2)->name);
 }
}
}

ped_person::~ped_person() {;}

void linksys_pedigree::show_loci(dcwindow *w)
{
if (w)
  {
  char buff[5];
  w->clear();
  for (int i=0;i<numloci;++i)
    w->printf("%2s %-4s %s%s",
//      loci[i].tofix?"4F":
      loci[i].type==QUAN?"Q":
#ifdef __ZTC__
      loci[i].type==AFF?"A":itoa(loci[i].nalleles,buff,10),
#else
      loci[i].type==AFF?"A":(sprintf(buff,"%d",loci[i].nalleles),buff),
#endif
      loci[i].name,loci[i].comment,(i<numloci-1)?"\n":"");
  }
else
  for (int i=0;i<numloci;++i)
    printf("%2d %4s %s",loci[i].nalleles,loci[i].name,loci[i].comment);
}

locus& locus::operator=(const locus &old)
{
nliab=old.nliab;
male_nliab=old.male_nliab;
chromosome=old.chromosome;
//affection=old.affection;
//tofix=old.tofix;
tocondense=old.tocondense;
//quan=old.quan;
type=old.type;
q_AA=old.q_AA;
q_aa=old.q_aa;
q_Aa=old.q_Aa;
q_var=old.q_var;
het_ratio=old.het_ratio;
nalleles=old.nalleles;
strcpy(comment,old.comment);
strcpy(name,old.name);
int i,j;
if (maxpen<nliab)
  {
  maxpen=nliab;
  if (pen) free((char*)pen);
  if (male_pen) free((char*)male_pen);
#ifdef SUNCC
  male_pen=(double2 *)malloc(maxpen*2*sizeof(double));
#else
  male_pen=malloc(maxpen*2*sizeof(double));
#endif
#ifdef SUNCC
  if ((pen=(double3 *)malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#else
  if ((pen=malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#endif
    { free((char*)male_pen); male_pen=0; }
  if (!pen) { dcerror(ENOMEM); maxpen=0; }
  }
for (i=0;i<nliab;++i)
  {
  for (j=0;j<3;++j)
     {
     pen[i][j]=old.pen[i][j];
     if (j<2) male_pen[i][j]=old.male_pen[i][j];
     }
  }
if (maxall<nalleles)
  {
  if (allele) delete[] allele;
  maxall=nalleles;
  allele=new allele_t[maxall];
  if (!allele)
     { dcerror(ENOMEM); maxall=0; }
  }
for (i=0;i<nalleles;++i)
  {
  allele[i].freq=old.allele[i].freq;
  strcpy(allele[i].marker,old.allele[i].marker);
  strcpy(allele[i].bin,old.allele[i].bin);
  if (old.allele[i].all_comment && (allele[i].all_comment=strdup(old.allele[i].all_comment))==0)
       dcerror(ENOMEM,"Out of memory in locus::operator=()");
  }
return *this;
}

locus& locus::shrink_copy(const locus &old)
{
nliab=old.nliab;
male_nliab=old.male_nliab;
chromosome=old.chromosome;
//affection=old.affection;
//tofix=old.tofix;
tocondense=old.tocondense;
//quan=old.quan;
type=old.type;
q_AA=old.q_AA;
q_aa=old.q_aa;
q_Aa=old.q_Aa;
q_var=old.q_var;
het_ratio=old.het_ratio;
nalleles=old.nalleles;
strcpy(comment,old.comment);
strcpy(name,old.name);
int i,j;
if (maxpen!=nliab)
  {
  maxpen=nliab;
  if (pen) free((char*)pen);
  if (male_pen) free((char*)male_pen);
  if (nliab==0)
    { pen=0; male_pen=0; }
  else
    {
#ifdef SUNCC
    male_pen=(double2 *)malloc(maxpen*2*sizeof(double));
#else
    male_pen=malloc(maxpen*2*sizeof(double));
#endif
#ifdef SUNCC
    if ((pen=(double3 *)malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#else
    if ((pen=malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#endif
      { free((char*)male_pen); male_pen=0; }
    if (!pen) { dcerror(ENOMEM); maxpen=0; }
    }
  }
for (i=0;i<nliab;++i)
  {
  for (j=0;j<3;++j)
     {
     pen[i][j]=old.pen[i][j];
     if (j<2) male_pen[i][j]=old.male_pen[i][j];
     }
  }
if (maxall!=nalleles)
  {
  if (allele) delete[] allele;
  maxall=nalleles;
  if (maxall==0)
    allele=0;
  else
    {
    allele=new allele_t[maxall];
    if (!allele)
      { dcerror(ENOMEM); maxall=0; }
    }
  }
for (i=0;i<nalleles;++i)
  {
  allele[i].freq=old.allele[i].freq;
  strcpy(allele[i].marker,old.allele[i].marker);
  strcpy(allele[i].bin,old.allele[i].bin);
  if (old.allele[i].all_comment && (allele[i].all_comment=strdup(old.allele[i].all_comment))==0)
       dcerror(ENOMEM,"Out of memory in locus::shrink_copy()");
  }
return *this;
}

allele_t::~allele_t()
{ if (all_comment) free(all_comment); }

allele_t::allele_t()
{ all_comment=0; }

locus::~locus()
{
if (allele) delete[] allele;
if (pen) free((char*)pen);
if (male_pen) free((char*)male_pen);
}

locus::locus(int ma,int mp)
{
create(ma,mp);
}

locus::locus()
{
create(MAXALL,MAXLIAB);
}

void locus::create(int ma,int mp)
{
type=MARKER;
maxall=ma;
maxpen=mp;
if (maxall==0)
  allele=0;
else
  {
  allele=new allele_t[maxall];
  if (!allele)
    { dcerror(ENOMEM); maxall=0; }
  }
if (maxpen==0)
  { pen=0; male_pen=0; }
else
  {
#ifdef SUNCC
  male_pen=(double2 *)malloc(maxpen*2*sizeof(double));
#else
  male_pen=malloc(maxpen*2*sizeof(double));
#endif
#ifdef SUNCC
  if ((pen=(double3 *)malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#else
  if ((pen=malloc(maxpen*3*sizeof(double)))==0 && male_pen)
#endif
     { free((char*)male_pen); male_pen=0; }
  if (!pen) { dcerror(ENOMEM); maxpen=0; }
  }
clear();
}

void locus::clear()
{
// affection=quan=0;
nliab=male_nliab=chromosome=nalleles=tocondense=0; // tofix=0
q_AA=q_aa=q_Aa=q_var=0.0;
het_ratio=1.0;
type=MARKER;
comment[0]=name[0]='\0';
int i,j;
for (i=0;i<maxpen;++i)
  {
  for (j=0;j<3;++j)
     {
     pen[i][j]=0.0;
     if (j<2) male_pen[i][j]=0.0;
     }
  }
for (i=0;i<maxall;++i)
  {
  allele[i].freq=0.0;
  allele[i].bin[0]=allele[i].marker[0]='\0';
  if (allele[i].all_comment)
    {
    free(allele[i].all_comment);
    allele[i].all_comment=0;
    }
  }
}

int linksys_pedigree::join_loops()
{
if (pointers_done) clear_pointers();
if (loops_broken && strcmp(member(count-1)->group,"0"))
  {
  int i,j,c;
  for (i=0;i<count;++i)
    {
    if ((c=atoi(member(i)->group))>1)
       {
       for (j=i+1;j<count;++j)
         if (atoi(member(j)->group)==c)
           break;
       if (j==count) // no match found
         if (c==2)
           {
           for (j=i+1;j<count;++j)
             if (atoi(member(j)->group)==1)
             {
             proband=member((i<j)?i:j);
             strcpy(proband->group,"9");
             }
           }
         else
           dcerror(3,"Tried to rejoin loops but could not match %s",(char*)member(i)->id);
       member(i)->id.move(member(j)->id);
       remove(j);
       }
    }
  }
if (proband && !users_proband)
  {
  strcpy(proband->group,proband->group[0]=='9'?"2":"0");
  proband=0;
  }
loops_broken=0;
return 1;
}

int linksys_pedigree::break_loops()
{
linksys_person *temp,*lsptr;
int c,i;
if (loops_broken) return 1;
if (!proband)
  {
  for (i=0;i<count;++i)
     if (!strcmp(member(i)->group,"1")||
         !strcmp(member(i)->group,"9"))
       {
       proband=member(i);
       users_proband=1;
       break;
       }
  if (i==count)
     {
     if (!pointers_done && !do_pointers()) return 0;
     users_proband=0;
     for (i=0;i<count;++i)
      if (member(i)->get_father() && !member(i)->numspo)
        {
        proband=(linksys_person*)member(i)->find_proband();
        int c;
        if ((c=atoi(proband->group))==2)
          strcpy(proband->group,"9");
        else if (c)
          {
          dcerror(2,"Automatically selected proband is in a loop");
          return 0;
          }
        else
          strcpy(proband->group,"1");
        break;
        }
     }
  }
if (pointers_done) clear_pointers();
int count_now=count;
for (i=0;i<count_now;++i)
  {
  lsptr=member(i);
  if ((c=atoi(lsptr->group))>1)
    {
    temp=new linksys_person;
    if (dcerror.stat()) return 0;
    *temp=*lsptr;
    if (c==9)
      {
      strcpy(lsptr->group,"1");
      strcpy(temp->group,"2");
      }
    temp->fid.strfree();
    temp->mid.strfree();
    char buffer[31];
    sprintf(buffer,"XX%s",(char*)lsptr->id);
    if (!lsptr->id.assign(buffer))
      dcerror(ENOMEM,"Out of memory in linksys_pedigree::break_loops()");
    add_member((person*)temp);
    if (dcerror.stat()) return 0;
    }
  }
loops_broken=1;
return 1;
}

int locus::output(data_file &file)
{
#if 0
fprintf(get_fp(file),"%s %s%s%s",name,tofix?"FIX ":tocondense?"CONDENSE ":quan?"Q ":"",
          comment,NL_STR);
#endif
fprintf(get_fp(file),"%s %s%s\n",name,tocondense?"CONDENSE ":type==QUAN?"Q ":"",
          comment);
for (int i=0;i<nalleles;++i)
  {
  if (type==AFF || type==QUAN)
    fprintf(get_fp(file),"%g\n",allele[i].freq);
  else
    fprintf(get_fp(file),"%g %s %s %s\n",allele[i].freq,
       (type==AFF || type==QUAN)?"":allele[i].marker,
       type==BINARY?allele[i].bin:"",
       (type==AFF || type==QUAN)?"":allele[i].all_comment?
         allele[i].all_comment:"");
  }
if (type==QUAN)
  fprintf(get_fp(file),"%g %g %g\n%g %g\n",
     q_AA,q_Aa,q_aa,q_var,het_ratio);
else if (type==AFF)
  {
  for (i=0;i<nliab;++i)
    fprintf(get_fp(file),"%g %g %g\n",
        pen[i][0],
        pen[i][1],
        pen[i][2]);
  for (i=0;i<nliab;++i)
    fprintf(get_fp(file),"%g %g\n",
        male_pen[i][0],
        male_pen[i][1]);
  }
fprintf(get_fp(file),"\n");
return 1;
}

static locus temp_input;

int locus::input(data_file &file)
{
int rv,old_error=dcerror.stat();
rv=temp_input.real_input(file);
shrink_copy(temp_input);
return rv && (old_error==dcerror.stat());
}
// I know it's pathetic, but it was quick to write


int locus::real_input(data_file &file)
{
char line[241],buff[21],rest[241];
clear();
do    {					 // read in initial blank or bad lines
      *buff='\0';
      if (!fgets(line,240,get_fp(file)))
         return 0;
      if (sscanf(line," %20s",buff)!=1) continue;
      if (!stricmp(buff,"END")) return 0;
      } while (!isalnum(*buff) && !isprint(*buff));
if (sscanf(line," %4s %80[^\n]",name,comment)==2)
  {
  *rest='\0';
  if (sscanf(comment," %10s %[^\n]",buff,rest)>=1)
    {
    if (!strcmp(strupr(buff),"Q"))
      { type=QUAN; strcpy(comment,rest); }
    else if (!strcmp(strupr(buff),"CONDENSE"))
      { tocondense=1; strcpy(comment,rest); }
    }
  if (sscanf(comment," %5s ",buff)==1 && toupper(buff[0])=='C')
    {
    if (toupper(buff[1])=='X') chromosome=-1;
    else chromosome=atoi(buff+1);
    }
  }
for (nalleles=0;
     nalleles<MAXALL && fgets(line,240,get_fp(file))
       && !(type==QUAN && nalleles>1); ++nalleles)
  {
  char commbuff[65];
  int c;
  if ((c=sscanf(line," %lf %20s %65[^\n]",
      &allele[nalleles].freq,
      allele[nalleles].marker,commbuff)) <1)  break;
  if (c==1)
    { if (type!=QUAN) type=AFF; }
  else if (c==3)
    {
    if (type==AFF) break; // onto penetrances now
    *rest='\0';
    sscanf(commbuff,"%20s %[^\n]",allele[nalleles].bin,rest);
    if (strspn(allele[nalleles].bin,"10")==strlen(allele[nalleles].bin))
      { strcpy(commbuff,rest); type=BINARY; }
    else allele[nalleles].bin[0]='\0';
    if (strlen(commbuff) && (allele[nalleles].all_comment=strdup(commbuff))==0)
      dcerror(ENOMEM,"Out of memory in locus::input()");
    }
  }
if (type==QUAN)
  {
  sscanf(line,"%lf %lf %lf ",&q_AA,&q_Aa,&q_aa);
  fgets(line,240,get_fp(file));
  het_ratio=1.0;
  sscanf(line,"%lf %lf",&q_var,&het_ratio);
  }
else if (type==AFF)
  {
  int c;
#if 0
  strcpy(allele[0].marker,"1");
  strcpy(allele[1].marker,"2");
#endif
  do
    {
    c=sscanf(line," %lf %lf %lf ",
        &pen[nliab][0],
        &pen[nliab][1],
        &pen[nliab][2]);
    if (c==3)
      {
      male_pen[nliab][0]=pen[nliab][0];
      male_pen[nliab][1]=pen[nliab][2];
      ++nliab;
      }
    } while (c==3 && fgets(line,240,get_fp(file)));
  if (c==2) // male penetrances
   {
   pen[nliab][0]=pen[nliab][1]=0.0;
   do
    {
    c=sscanf(line," %lf %lf ",
        &male_pen[male_nliab][0],
        &male_pen[male_nliab][1]);
    if (c==2) ++male_nliab;
    } while (c==2 && fgets(line,240,get_fp(file)));
   }
  male_nliab=nliab;
  }
else if (type==BINARY)
  {
  int nbin=0;
  for (int i=0;i<nalleles;++i)
    nbin+=(allele[i].bin[0]!=0);
  if (nbin && nbin!=nalleles)
    {
    dcerror(3,"Only %d binary marker%s supplied for locus %s\n",
       nbin,nbin>1?"s":"",name);
    return 0;
    }
  }
return 1;
}

int locus::to_datafile(FILE *fp,int sex_linked,int strip_labels)
{
int i;
if (strip_labels)
  fprintf(fp," %d %d%s",type,nalleles,NL_STR);
else
  fprintf(fp," %d %d  << %s, #alleles [%4.4s]%s",
    type,nalleles,
    type==QUAN?"quantitative":
       type==AFF?"affection":
       type==BINARY?"binary factors":"numbered alleles",
    name,NL_STR);
for (i=0;i<nalleles;++i)
    fprintf(fp," %E  ",allele[i].freq);
fprintf(fp,"%s%s",strip_labels?"":"<< gene freqs",NL_STR);
  if (type==QUAN)
    {
    fprintf(fp," 1 %s%s",
      strip_labels?"":"<< number of quantitative variables",NL_STR);
    fprintf(fp," %7.5f %7.5f %7.5f %s%s",
      q_AA,q_Aa,q_aa,strip_labels?"":"<< means",NL_STR);
    fprintf(fp," %7.5f %s%s",q_var,strip_labels?"":"<< variance",NL_STR);
    fprintf(fp," %7.5f %s%s",het_ratio,strip_labels?"":"<< multiplier for heterozygote variance",
       NL_STR);
    }
  else if (type==AFF)
    {
    int liab;
    fprintf(fp," %d %s%s",nliab,
        strip_labels?"":"<< number of liability classes",NL_STR);
    for (liab=0;liab<nliab;++liab)
      {
      fprintf(fp," %7.5f %7.5f %7.5f%s",
      pen[liab][0],pen[liab][1],pen[liab][2],NL_STR);
      if (chromosome==-1  || sex_linked)
        fprintf(fp," %7.5f %7.5f%s",
         nliab==male_nliab?male_pen[liab][0]:pen[liab][0],
         nliab==male_nliab?male_pen[liab][1]:pen[liab][2],NL_STR);
      }
    }
  else if (type==BINARY)
    {
    fprintf(fp," %d  %s%s",
             strlen(allele[i].bin)?strlen(allele[i].bin):nalleles,
             strip_labels?"":"<< number of binary factors",
             NL_STR);
    int k;
    for (i=0;i<nalleles;++i)
      if (!strlen(allele[i].bin))
        for (k=0;k<nalleles;++k)
          fprintf(fp,"%c%s",(i==k)?'1':'0',
            (k==nalleles-1)?NL_STR:" ");
      else
        for (k=0;k<strlen(allele[i].bin);++k)
          fprintf(fp,"%c%s",
            allele[i].bin[k],
              (k==nalleles-1)?NL_STR:" ");
    }
fprintf(fp,NL_STR);
return 1;
}


int linksys_pedigree::sort_markers(char *fn,locus **gl)
{
char name[65];
if (!*fn) //default
  {
  strcpy(name,getenv("LOCUSLIB")?getenv("LOCUSLIB"):"LOCUSDAT.DAT");
  fn=name;
  }
hash_file locusdat;
locus temp;
locusdat.open(fn,"r");
numloci=member(0)->nloci;
if (*gl) { delete[] *gl; *gl=NULL; }
if (numloci && (*gl=new locus[numloci])==NULL) return dcerror(ENOMEM);
loci=*gl;
for (int i=0;i<numloci;++i)
  {
  if (locusdat.fill(&temp,strupr(member(0)->locphen[i].loc)))
       loci[i]=temp;
  }
return numloci;
}

int linksys_pedigree::output_par(char *fn,int loc_to_do,char locname[][5],
  ls4parms &lp)
{
char *ptr;
if ((ptr=strchr(fn,'.'))!=0) *ptr='\0';
strcat(fn,".par");
FILE *fp=fopen(fn,"wb");
int i,j,loc[MAXLOCI];
for (j=0;j<loc_to_do;++j)
  {
  loc[j]=-1;
  for (i=0;i<numloci;++i)
    if (!strcmp(loci[i].name,locname[j]))
      {
      loc[j]=i;
      break;
      }
  if (loc[j]==-1)
     dcerror(1,"Unknown locus %s",locname[j]);
  }
if (dcerror.stat()) return 0;
int sex_linked=0;
for (j=0;j<loc_to_do;++j)
  if (loci[loc[j]].chromosome==-1) sex_linked=1;
if (lp.strip_labels)
  {
  fprintf(fp," %d   0   %d   5 %s",loc_to_do,
    sex_linked,NL_STR);
  fprintf(fp," 0  0.0  0.0  0 %s",NL_STR);
  for (i=1;i<=loc_to_do;++i)
    fprintf(fp," %d",i);
  fprintf(fp," %s%s",NL_STR,NL_STR);
  }
else
  {
  fprintf(fp," %d   0   %d   %d  << no loci, risk locus, sexlinked(if 1)%s",loc_to_do,
    sex_linked,
    5, // change this to be correct for program
    NL_STR);
  fprintf(fp," 0  0.0  0.0  0  << mut locus, mut rate, haplotype freq(if 1)%s",NL_STR);
  for (i=1;i<=loc_to_do;++i)
    fprintf(fp," %d",i);
  fprintf(fp," << order of loci%s%s",NL_STR,NL_STR);
  }
for (j=0;j<loc_to_do;++j)
  loci[loc[j]].to_datafile(fp,sex_linked,lp.strip_labels);
fclose(fp);
return loci[loc[1]].chromosome;
}

int linksys_pedigree::output_mflink_par(char *fn,int loc_to_do,char locname[][5],float rf[],ls4parms &lp)
{
char *ptr;
if ((ptr=strchr(fn,'.'))!=0) *ptr='\0';
strcat(fn,".uda");
FILE *fp[2];
fp[0]=fopen(fn,"wb");
if ((ptr=strchr(fn,'.'))!=0) *ptr='\0';
strcat(fn,".lda");
fp[1]=fopen(fn,"wb");
int i,j,loc[MAXLOCI];
for (j=0;j<loc_to_do;++j)
  {
  loc[j]=-1;
  for (i=0;i<numloci;++i)
    if (!strcmp(loci[i].name,locname[j]))
      {
      loc[j]=i;
      break;
      }
  if (loc[j]==-1)
     dcerror(1,"Unknown locus %s",locname[j]);
  }
int aff=-1;
for (j=0;j<loc_to_do;++j)
  {
  if (loci[loc[j]].type==AFF)
    if (aff!=-1) dcerror(1,"More than one affection locus: %s",locname[j]);
    else aff=j;
  }
if (aff==-1) dcerror(1,"No affection locus for MFLINK analysis");
if (dcerror.stat()) return 0;
int sex_linked=0;
for (j=0;j<loc_to_do;++j)
  if (loci[loc[j]].chromosome==-1) sex_linked=1;
for (int f=0;f<2;++f)
  {
  fprintf(fp[f]," %d   0   %d   5  << no loci, risk locus, sexlinked(if 1)%s",loc_to_do,
    sex_linked,NL_STR);
  fprintf(fp[f]," 0  0.0  0.0  0  << mut locus, mut rate, haplotype freq(if 1)%s",NL_STR);
  if (f==1)
    for (i=1;i<=loc_to_do;++i) fprintf(fp[f]," %d",i);
  else
    {
    fprintf(fp[f],"%d ",aff+1);
      for (i=1;i<=loc_to_do;++i) if (i!=aff+1) fprintf(fp[f]," %d",i);
    }
  fprintf(fp[f]," << order of loci%s%s",NL_STR,NL_STR);
for (j=0;j<loc_to_do;++j)
  loci[loc[j]].to_datafile(fp[f],sex_linked,lp.strip_labels);
fprintf(fp[f],"0 0%s",NL_STR);
  if (f==1)
    for (i=0;i<loc_to_do-1;++i) fprintf(fp[f]," %f",rf[i]);
  else
    {
    fprintf(fp[f]," 0.5");
    for (i=0;i<loc_to_do-1;++i)
      if (i!=aff) fprintf(fp[f]," %f",(i<aff-1||i>aff)?rf[i]:rf[i]+rf[i+1]);
    }
  fprintf(fp[f],"%s1 2 1%s",NL_STR,NL_STR);
  fclose(fp[f]);
  }
return 1;
}

int ped_person::set_bin(int i,char *s,locus_type t)
{
bin[i].type=t;
return bin[i].value.assign(s);
}

int locus::which_allele(char c)
{
for (int k=0;k<nalleles;++k)
      if (strchr(allele[k].marker,c))
         return k;
return -1;
}

int linksys_pedigree::output_ped(int fnum,data_file &file,int loc_to_do,char locname[][5],
  ls4parms &lp)
{
int i,j,l,loc[MAXLOCI];
int write_condkey=lp.condkey;
int do_ppd=lp.do_ppd;
int strip_labels=lp.strip_labels;
for (j=0;j<loc_to_do;++j)
  {
  loc[j]=-1;
  for (i=0;i<numloci;++i)
    if (!strcmp(loci[i].name,locname[j]))
      {
      loc[j]=i;
      if (loci[i].tocondense)
       {
       data_file fkey;
       if (write_condkey)
         {
         fkey.open("CONDENSE.KEY","a");
         fkey.fprintf("%s, ",loci[i].name);
         }
       if (!condense(i,loci[i].nalleles,write_condkey?&fkey:0)) return 0;
       }
      break;
      }
  if (loc[j]==-1) dcerror(1,"Unkown locus %s",locname[j]);
  }
if (dcerror.stat()) return 0;
// assume all sorted
linksys_person *lsptr;
ped_person *temp;
for (i=0;i<count;++i)
  {
  temp=do_ppd?new ppd_person:new ped_person;
  if (temp==0) return(ENOMEM);
  lsptr=member(i);
  temp->from_linksys(*lsptr);
  if (dcerror.stat()) { delete temp; return 0; }
  if (fnum!=-1) temp->set_famnum(fnum);
  for (l=0;l<loc_to_do;++l)
    {
    if (loci[loc[l]].type==AFF)
      {
      if (loc[l]>=lsptr->nloci ||
          !*lsptr->locphen[loc[l]].phen) // blank
            temp->set_bin(l,loci[loc[l]].nliab>1?"01":"0",AFF);
      else
          temp->set_bin(l,lsptr->locphen[loc[l]].phen,AFF);
      }
    else if (loci[loc[l]].type==QUAN)
      {
      if (loc[l]>=lsptr->nloci ||
          !*lsptr->locphen[loc[l]].phen) // blank
          temp->set_bin(l,"0.0",QUAN);
      else
          temp->set_bin(l,lsptr->locphen[loc[l]].phen,QUAN);
      }
    else if (loci[loc[l]].type==BINARY)
      {
      int j;
      char tempbin[21];
      tempbin[0]='\0';
      for (j=0;j<loci[loc[l]].nalleles;++j)
         strcat(tempbin,"0");
      if (loc[l]<lsptr->nloci)
          for (char *ptr=lsptr->locphen[loc[l]].phen+(loci[loc[l]].tocondense)*6;
                  *ptr;++ptr)
        if (isalnum(*ptr) && *ptr!='0')
        {
        if ((j=loci[loc[l]].which_allele(*ptr))!=-1)
            {
            if (strlen(loci[loc[l]].allele[j].bin))
              for (int k=0;k<strlen(loci[loc[l]].allele[j].bin);++k)
                tempbin[k]=max(tempbin[k],loci[loc[l]].allele[j].bin[k]);
            else
              tempbin[j]='1';
            }
        else
          {
          dcerror(2,"%s: %c is not in list of markers for locus %s\n",
             (char*)lsptr->id,*ptr,loci[loc[l]].name);
          return 0;
          }
        }
      temp->set_bin(l,tempbin,BINARY);
      }
    else
      {
      int alls[2],done;
      char tempbin[10];
      strcpy(tempbin,"00");
      alls[0]=alls[1]=done=0;
      if (loc[l]<lsptr->nloci)
        for (char *ptr=lsptr->locphen[loc[l]].phen+
         (loci[loc[l]].tocondense)*6;
              *ptr;++ptr)
        if (isalnum(*ptr) && *ptr!='0')
        {
        if ((j=loci[loc[l]].which_allele(*ptr))!=-1)
            {
            if (done<2)
              alls[done++]=1+j;
            else
              {
              dcerror(3,"Too many alleles in %s - %s",(char*)lsptr->id,lsptr->locphen[loc[l]].phen);
              return 0;
              }
            }
        else
          {
          dcerror(2,"%c is not in list of markers for locus %s\n",
             *ptr,loci[loc[l]].name);
          return 0;
          }
        }
      sprintf(tempbin,"%2d %-2d",alls[0],done!=1?alls[1]:alls[0]);
      temp->set_bin(l,tempbin,MARKER);
      }
    }
  temp->set_numloci(loc_to_do);
  temp->output(file,strip_labels);
  delete temp;
  }
return 1;
}


int linksys_pedigree::output_bat(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkpath,
  int use_mlink)
{
char lpath[70];
strcpy(lpath,linkpath);
if (*lpath && lpath[strlen(lpath)-1] != '\\')
   strcat(lpath,"\\");
FILE *fp=fopen(fn,"w");
int i,j;
*strchr(fn,'.')='\0';
fprintf(fp,"ECHO OFF\n\
BREAK ON\n\
DEL SETUP.LOG\n\
DEL STREAM.DAT\n\
DEL FINAL.DAT\n\
DEL OUTFILE.DAT\n\
\n\
REM IF EXIST %s.RES COPY %s.RES %s.BAK\n\
TYPE NUL > %s.RES\n\
\n\
REM IF EXIST %s.OUT COPY %s.OUT %s.BAK\n\
TYPE NUL > %s.OUT\n\
\n\
COPY %s.PPD PPPEDIN.DAT\n\
COPY %s.PAR PPDATAIN.DAT\n\
\n",
fn,fn,fn,fn,fn,fn,fn,fn,fn,fn);
if (loc_to_do==2 || use_mlink) // MLINK
  {
  fprintf(fp,"ECHO. \n\
ECHO. \n\
ECHO **************************************************************************\n\
ECHO Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,"\nECHO Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,">> %s.RES\n\
ECHO **************************************************************************\n\
ECHO. \n\
ECHO MLINK PPPEDIN.DAT PPDATAIN.DAT SETUP.LOG %d 0 1 2 1 7 > $$$SETUP.TMP\n",
  fn,loc_to_do);
fprintf(fp,"ECHO ");
for (i=0;i<loc_to_do;++i)  fprintf(fp,"%d ",i+1);
fprintf(fp,">> $$$SETUP.TMP\n\
ECHO 0 ");
for (i=0;i<loc_to_do-2;++i)  fprintf(fp,"%6.4f ",rf[i]);
fprintf(fp,
">> $$$SETUP.TMP\n\
ECHO 0.001 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.01 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.05 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.1 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.2 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.3 2.0 1.0 >> $$$SETUP.TMP\n\
ECHO 0.4 2.0 1.0 >> $$$SETUP.TMP\n\
%sSETUP @$$$SETUP.TMP\n\
IF ERRORLEVEL 1 GOTO L1\n\
TYPE SETUP.LOG >> %s.RES\n\
TYPE STREAM.DAT >> %s.OUT\n\
IF \"%%1\" == \"NODELETE\" GOTO L0\n\
DEL $$$SETUP.TMP\n\
DEL SETUP.LOG\n\
DEL STREAM.DAT\n\
:L0\n\
%sUNKNOWN\n\
IF ERRORLEVEL 2 GOTO L1\n\
IF ERRORLEVEL 1 GOTO LEXIT\n\
%sMLINK\n\
IF ERRORLEVEL 2 GOTO L1\n\
IF ERRORLEVEL 1 GOTO LEXIT\n\
TYPE OUTFILE.DAT >> %s.RES\n\
TYPE STREAM.DAT >> %s.OUT\n\
IF \"%%1\" == \"NODELETE\" GOTO L1\n\
DEL IPEDFILE.DAT\n\
DEL PEDFILE.DAT\n\
DEL DATAFILE.DAT\n\
DEL SPEEDFILE.DAT\n\
DEL OUTFILE.DAT\n\
DEL STREAM.DAT\n\
:L1\n\
\n\
:LEXIT\n\
\n\
IF \"%%1\" == \"NODELETE\" GOTO LEND\n\
DEL STREAM.DAT\n\
DEL IPEDFILE.DAT\n\
DEL PEDFILE.DAT\n\
DEL DATAFILE.DAT\n\
DEL TEMPPED.DAT\n\
DEL TEMPDAT.DAT\n\
DEL DIAGFILE.DAT\n\
DEL SPEEDFILE.DAT\n\
DEL OUTFILE.DAT\n\
DEL FINAL.DAT\n\
DEL STREAM.DAT\n\
DEL $$$SETUP.TMP\n\
DEL SETUP.LOG\n\
DEL PPPEDIN.DAT\n\
DEL PPDATAIN.DAT\n\
DEL MPPEDIN.DAT\n\
DEL MPDATAIN.DAT\n\
DEL IPPEDIN.DAT\n\
DEL IPDATAIN.DAT\n\
DEL MIPEDIN.DAT\n\
DEL MIDATAIN.DAT\n\
:LEND\n",
  lpath,
  fn,fn,
  lpath,
  lpath,
  fn,fn);
  }
else
  {
  int label=0;
  for (i=0;i<loc_to_do;++i)
    {
    if (i>0 && i<loc_to_do-1 && rf[i-1]<1.5e-3) continue;
    fprintf(fp,"ECHO \n\
ECHO. \n\
ECHO **************************************************************************\n\
ECHO Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp,"\nECHO Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp," >> %s.RES\n",fn);
    fprintf(fp,"ECHO **************************************************************************\n\
ECHO. \n\
ECHO LINKMAP PPPEDIN.DAT PPDATAIN.DAT SETUP.LOG %d 1 1 %5.3f 5 > $$$SETUP.TMP\n\
ECHO. ",
    loc_to_do,
    (i==0)?0.0:(i==loc_to_do-1)?0.5:rf[i-1]);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0) fprintf(fp,"%d ",j+1);
      if (j==i) fprintf(fp,"1 ");
      }
    fprintf(fp,">> $$$SETUP.TMP\nECHO ");
    for (j=0;j<loc_to_do;++j)
      {
      if (j>1) fprintf(fp,"%5.3f ",rf[j-2]);
      if (j==i) fprintf(fp,"%s ",i?"0":"0.5");
      }
    fprintf(fp,">> $$$SETUP.TMP\nECHO %4.2f >> $$$SETUP.TMP\n\
%sSETUP @$$$SETUP.TMP\n\
IF ERRORLEVEL 1 GOTO L%d\n\
TYPE SETUP.LOG >> %s.RES\n\
TYPE STREAM.DAT >> %s.OUT\n\
IF \"%%1\" == \"NODELETE\" GOTO L%d\n\
DEL $$$SETUP.TMP\n\
DEL SETUP.LOG\n\
DEL STREAM.DAT\n\
:L%d\n",
    sexratio,
    lpath,
    label+1,fn,fn,label,label);
    fprintf(fp,"%sUNKNOWN\n\
IF ERRORLEVEL 2 GOTO L%d\n\
IF ERRORLEVEL 1 GOTO LEXIT\n\
%sLINKMAP\n\
IF ERRORLEVEL 2 GOTO L%d\n\
IF ERRORLEVEL 1 GOTO LEXIT\n\
TYPE OUTFILE.DAT >> %s.RES\n\
TYPE STREAM.DAT >> %s.OUT\n\
IF \"%%1\" == \"NODELETE\" GOTO L%d\n\
DEL IPEDFILE.DAT\n\
DEL PEDFILE.DAT\n\
DEL DATAFILE.DAT\n\
DEL SPEEDFILE.DAT\n\
DEL OUTFILE.DAT\n\
DEL STREAM.DAT\n\
:L%d\n\n",
    lpath,
    label+1,
    lpath,
    label+1,fn,fn,label+1,label+1);
    label+=2;
    }
  fprintf(fp,":LEXIT\n\
\n\
IF \"%%1\" == \"NODELETE\" GOTO LEND\n\
DEL STREAM.DAT\n\
DEL IPEDFILE.DAT\n\
DEL PEDFILE.DAT\n\
DEL DATAFILE.DAT\n\
DEL TEMPPED.DAT\n\
DEL TEMPDAT.DAT\n\
DEL DIAGFILE.DAT\n\
DEL SPEEDFILE.DAT\n\
DEL OUTFILE.DAT\n\
DEL FINAL.DAT\n\
DEL STREAM.DAT\n\
DEL $$$SETUP.TMP\n\
DEL SETUP.LOG\n\
DEL PPPEDIN.DAT\n\
DEL PPDATAIN.DAT\n\
DEL MPPEDIN.DAT\n\
DEL MPDATAIN.DAT\n\
DEL IPPEDIN.DAT\n\
DEL IPDATAIN.DAT\n\
DEL MIPEDIN.DAT\n\
DEL MIDATAIN.DAT\n\
:LEND");
  }
fclose(fp);
return 1;
}

int linksys_pedigree::output_lsp_script(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkpath,
  int use_mlink,int sex_linked)
{
char lpath[70];
strcpy(lpath,linkpath);
if (*lpath && lpath[strlen(lpath)-1] != '/')
   strcat(lpath,"/");
FILE *fp=fopen(fn,"wb");
int i,j;
*strchr(fn,'.')='\0';
fprintf(fp,
"rm -f setup.log stream stream.dat final final.dat final \\\012\
    pedfile pedfile.dat pedfile ipedfile.dat outfile.dat outfile \\\012\
    speedfile speedfile.dat %s.res %s.out \012\012",fn,fn);
fprintf(fp,
"cp %s.ppd pppedin \012\
cp %s.par ppdatain \012\012",fn,fn);
if (loc_to_do==2 || use_mlink) // MLINK
  {
  fprintf(fp,
"echo \"**************************************************************************\"\012\
echo Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,"\012 echo Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,">> %s.res\012\
echo \"**************************************************************************\"\012\
echo \012",fn);
fprintf(fp,"lsp mlink pppedin ppdatain %d \\\012   ",loc_to_do);
for (i=0;i<loc_to_do;++i)  fprintf(fp,"%d ",i+1);
fprintf(fp,sexratio==1.0?"\\\012   0 0 ":"\\\012   1 0 ");
for (i=0;i<loc_to_do-2;++i)  fprintf(fp,"%6.4f ",rf[i]);
if (sexratio==1.0)
   fprintf(fp,"1 2 1 7 \\\012");
else
   fprintf(fp,"1 %6.4f 2 1 7 \\\012",sexratio);
fprintf(fp,
"0.001 2.0 1.0 \\\012\
0.01 2.0 1.0 \\\012\
0.05 2.0 1.0 \\\012\
0.1 2.0 1.0 \\\012\
0.2 2.0 1.0 \\\012\
0.3 2.0 1.0 \\\012\
0.4 2.0 1.0 \012\
if [ $? = '0' -o $? = '1' ]\012\
then\012\
   cat lsp.log >> %s.res\012\
   cat lsp.stm >> %s.out\012",fn,fn);
fprintf(fp,"   unknown\012   if [ $? = '0' ]\012   then\012\
   mlink\012   if [ $? = '0' ]\012   then\012\
     cat outfile.dat >> %s.res\012\
     cat stream.dat >> %s.out\012\
     fi\012\
     fi\012\
fi\012\012",fn,fn);
  }
else
  {
  for (i=0;i<loc_to_do;++i)
    {
    if (i>0 && i<loc_to_do-1 && rf[i-1]<1.5e-3) continue;
    fprintf(fp,
"echo \"**************************************************************************\"\012\
echo Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp,"\012echo Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp," >> %s.res\012",fn);
    fprintf(fp,"echo \"**************************************************************************\"\012\
echo \012");
fprintf(fp,"lsp linkmap %s ppdatain \\\012  %d \\\012",
    i==0?"pppedin":"=",
    loc_to_do);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0) fprintf(fp,"%d ",j+1);
      if (j==i) fprintf(fp,"1 ");
      }
    fprintf(fp,"\\\012  1  %5.3f 5 %d \\\012   ",
       (i==0)?0.0:(i==loc_to_do-1)?0.5:rf[i-1],1);
//       (i==0)?0.0:(i==loc_to_do-1)?0.5:rf[i-1],sex_linked);
    for (j=0;j<loc_to_do;++j)
      {
      if (j>1) fprintf(fp,"%5.3f ",rf[j-2]);
      if (j==i) fprintf(fp,"%s ",i?"0":"0.5");
      }
fprintf(fp,"\\\012  %5.3f \012",sexratio);
fprintf(fp,"\
if [ $? = '0' -o $? = '1' ]\012\
then\012\
   cat lsp.log >> %s.res\012\
   cat lsp.stm >> %s.out\012",fn,fn);
if (i==0)
  {
  fprintf(fp,"   unknown\012   if [ $? = '0' ]\012   then\012");
  }
fprintf(fp,"   linkmap\012   if [ $? = '0' ]\012   then\012\
     cat outfile.dat >> %s.res\012\
     cat stream.dat >> %s.out\012\
     fi\012\
%s\
fi\012\012",fn,fn,i==0?"   fi\012":"");
    }
  }
fprintf(fp,
"if test \"$1\" != \"nodelete\" \012\
   then \012\
   rm -f stream pedfile ipedfile datafile tempped tempdat diagfile speedfile outfile \\\012\
final stream pppedin ppdatain ippedin ipdatain mppedin mpdatain mipedin midatain \\\012\
pedfile.dat ipedfile.dat speedfile.dat datafile.dat outfile.dat stream.dat \\\012\
   setup.tmp setup.log \012\
fi \012");
fclose(fp);
return 1;
}

int linksys_pedigree::output_script(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkpath,
  int use_mlink,int sex_linked)
{
char lpath[70];
strcpy(lpath,linkpath);
if (*lpath && lpath[strlen(lpath)-1] != '/')
   strcat(lpath,"/");
FILE *fp=fopen(fn,"wb");
int i,j;
*strchr(fn,'.')='\0';
fprintf(fp,
"for fn in setup.log stream stream.dat final final.dat final \\\012\
    pedfile pedfile.dat pedfile ipedfile.dat outfile.dat outfile \\\012\
    speedfile speedfile.dat %s.res %s.out \012\
do \012\
if test -s $fn \012\
then rm $fn \012\
fi \012\
done \012\012",fn,fn);
fprintf(fp,
"cp %s.ppd pppedin \012\
cp %s.par ppdatain \012\012",fn,fn);
if (loc_to_do==2 || use_mlink) // MLINK
  {
  fprintf(fp,
"echo \"**************************************************************************\"\012\
echo Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,"\012 echo Run   1 - MLINK : ");
for (i=0;i<loc_to_do;++i) fprintf(fp,"%s ",locname[i]);
fprintf(fp,">> %s.res\012\
echo \"**************************************************************************\"\012\
echo \012\
echo mlink pppedin ppdatain setup.log %d 0 1 2 1 7 > setup.tmp\012",
  fn,loc_to_do);
fprintf(fp,"echo ");
for (i=0;i<loc_to_do;++i)  fprintf(fp,"%d ",i+1);
fprintf(fp,">> setup.tmp\012\
echo 0 ");
for (i=0;i<loc_to_do-2;++i)  fprintf(fp,"%6.4f ",rf[i]);
fprintf(fp,
">> setup.tmp\012\
echo 0.001 2.0 1.0 >> setup.tmp\012\
echo 0.01 2.0 1.0 >> setup.tmp\012\
echo 0.05 2.0 1.0 >> setup.tmp\012\
echo 0.1 2.0 1.0 >> setup.tmp\012\
echo 0.2 2.0 1.0 >> setup.tmp\012\
echo 0.3 2.0 1.0 >> setup.tmp\012\
echo 0.4 2.0 1.0 >> setup.tmp\012\
if %ssetup @setup.tmp\012\
   then \012\
   for fn in stream pedfile datafile \012\
   do \012\
   if test -s $fn \012\
   then cp $fn $fn.dat \012\
   fi \012\
   done \012\
   cat setup.log >> %s.res\012\
   cat stream.dat >> %s.out\012\
   if test \"$1\" != \"nodelete\" \012\
      then \012\
      rm setup.tmp\012\
      rm setup.log\012\
      rm stream.dat\012\
   fi \012\
   if %sunknown\012\
      then \012\
      for fn in ipedfile speedfile \012\
        do \012\
        if test -s $fn \012\
         then cp $fn $fn.dat \012\
        fi \012\
        done \012\
      if %smlink\012\
         then \012\
         for fn in outfile stream \012\
           do \012\
           if test -s $fn \012\
             then cp $fn $fn.dat \012\
           fi \012\
           done \012\
         cat outfile.dat >> %s.res\012\
         cat stream.dat >> %s.out\012\
      fi \012\
   fi \012\
fi \012",
  lpath,
  fn,fn,
  lpath,
  lpath,
  fn,fn);
  }
else
  {
  for (i=0;i<loc_to_do;++i)
    {
    if (i>0 && i<loc_to_do-1 && rf[i-1]<1.5e-3) continue;
    fprintf(fp,
"echo \"**************************************************************************\"\012\
echo Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp,"\012echo Run   %d - LINKMAP : ",i+1);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0)
        fprintf(fp,"%s ",locname[j]);
      if (j==i)
        fprintf(fp,"[%s] ",locname[0]);
      }
    fprintf(fp," >> %s.res\012",fn);
    fprintf(fp,"echo \"**************************************************************************\"\012\
echo \012\
echo linkmap pppedin ppdatain setup.log %d %d 1 %5.3f 5 > setup.tmp\012\
echo ",
    loc_to_do,
    sex_linked,
    (i==0)?0.0:(i==loc_to_do-1)?0.5:rf[i-1]);
    for (j=0;j<loc_to_do;++j)
      {
      if (j!=0) fprintf(fp,"%d ",j+1);
      if (j==i) fprintf(fp,"1 ");
      }
    fprintf(fp,">> setup.tmp\012echo ");
    for (j=0;j<loc_to_do;++j)
      {
      if (j>1) fprintf(fp,"%5.3f ",rf[j-2]);
      if (j==i) fprintf(fp,"%s ",i?"0":"0.5");
      }
    fprintf(fp,">> setup.tmp\012\
echo %4.2f >> setup.tmp\012\
if %ssetup @setup.tmp\012\
   then \012\
   for fn in stream pedfile datafile \012\
   do \012\
   if test -s $fn \012\
   then cp $fn $fn.dat \012\
   fi \012\
   done \012\
   cat setup.log >> %s.res\012\
   cat stream.dat >> %s.out\012\
   if test \"$1\" != \"nodelete\" \012\
      then \012\
      rm setup.tmp\012\
      rm setup.log\012\
      rm stream\012\
   fi \012",
sexratio,lpath,fn,fn);
fprintf(fp,
"   if %sunknown\012\
      then \012\
      for fn in ipedfile speedfile \012\
        do \012\
        if test -s $fn \012\
         then cp $fn $fn.dat \012\
        fi \012\
        done \012\
      if %slinkmap\012\
         then \012\
         for fn in outfile stream \012\
           do \012\
           if test -s $fn \012\
             then cp $fn $fn.dat \012\
           fi \012\
           done \012\
         cat outfile.dat >> %s.res\012\
         cat stream.dat >> %s.out\012\
      fi \012\
   fi \012\
fi \012",  // if setup
    lpath,lpath,fn,fn);
    }
  }
fprintf(fp,
"if test \"$1\" != \"nodelete\" \012\
   then \012\
   for fn in stream pedfile ipedfile datafile tempped tempdat diagfile speedfile outfile \\\012\
final stream pppedin ppdatain ippedin ipdatain mppedin mpdatain mipedin midatain \\\012\
pedfile.dat ipedfile.dat speedfile.dat datafile.dat outfile.dat stream.dat \\\012\
   setup.tmp setup.log \012\
   do \012\
   if test -s $fn \012\
      then rm $fn \012\
   fi \012\
   done \012\
fi \012");
fclose(fp);
return 1;
}


