/* PEDLINK.HPP */
/* Copyright Dave Curtis 1991 */
#ifndef PEDLINKHPP
#define PEDLINKHPP 1

#include "pedigree.hpp"
#include "dcstr.hpp"

#define MAXALL 20
// max number of alleles and binary factors per locus
#define MAXLIAB 20
// max number of liability classes per locus
#define MAXLOCI 100
// number of loci in par/ped/gen file
#define LOCPHENDIM 100
// number of loci in sdf file

enum locus_type { QUAN, AFF, BINARY, MARKER };

struct locphen_t
{
char loc[5],phen[9]; locus_type type; char decoded;
locphen_t();
};

// struct bincode_t { char value[11]; locus_type type; } ;
struct bincode_t { dcstring value; locus_type type; } ;

struct dc_locus_def_t;

class ped_person:public person
{
friend class tab_person;
protected:
int numloci;
dcstring famnum;
int idnum,fidnum,midnum;
bincode_t bin[MAXLOCI];
fill_markers(char *rest); // from datafile
fill_markers(char *rest,int n,dc_locus_def_t *def); // from datafile
public:
bincode_t get_bin(int i) { return bin[i]; }
ped_person &operator=(const ped_person &);
void set_famnum(int f)
{ char buff[10]; sprintf(buff,"%03d",f); famnum=buff; }
void set_numloci(int n) { numloci=n; }
int get_numloci() const { return numloci; }
set_bin(int i,char *s,locus_type t);
ped_person();
virtual ~ped_person();
virtual ped_person &from_linksys(const linksys_person& old);
virtual int input(data_file& file,int n,dc_locus_def_t *def);
virtual int input(data_file& file);
virtual int output(data_file& file,int strip_labels=0);
};

class ppd_person:public ped_person
{
friend class tab_person;
int psnum,msnum,offnum,probnum;
public:
ppd_person();
ped_person &from_linksys(const linksys_person& old);  // returns reference to self as ped_person
virtual int input(data_file& file);
virtual int input(data_file& file,int n,dc_locus_def_t *def);
virtual int output(data_file& file,int strip_labels=0);
};

class editable_linksys_person;

class editable_locus;
class locus;

class linksys_person:public person
{
friend class sdf_data_file;
friend class linksys_pedigree;
friend class criped;
friend class tab_person;
friend class ped_person;
friend class ppd_person;
friend class cri_person;
int ped_num;
char fname[9],
lsid[5],
group[2],
names[49],
age[7],
sex[2],
lsfid[5],
lsmid[5];
struct locphen_t *locphen;
int nloci;
public:
const char *get_fname() const { return fname; }
alloc_loci(int nl);
linksys_person();
~linksys_person();
//void *base() { return this; }
//#define linksys_person_ptr(x) ((x)?((linksys_person*)((x)->base())):0) // x is a person
//#define linksys_person_ptr(x) ((x)?((linksys_person*)(x)):0) // x is a person
revise(editable_linksys_person &lsp_ed);
virtual int input(data_file& file);
// in ranped.cpp:
int output(data_file& file,int nl=-1);
// not virtual, don't have to link it in
set_pen(int dloc,int aloc,float pen[3]);
randomise(int l1,int l2,float rf,locus *loci);
char add_to_buffer(char alls[5],int ch);
fix_one(int l,char alls[5]);
fix(int l,char alls[5]);
linksys_person &operator=(const linksys_person &old);
};

struct allele_t
{
double freq; char marker[11],*all_comment; char bin[21];
allele_t();
~allele_t();
};

extern "C" { int locus_compare(locus **l1,locus **l2); }

typedef double double2[2];
typedef double double3[3];

class locus : public data_object
{
friend int locus_compare(locus **l1,locus **l2);
friend class locus_data_file;
friend class linksys_pedigree;
friend class criped;
protected:
short maxall,maxpen;
locus_type type;
//char affection,quan
char tocondense; //,tofix;
double q_AA,q_aa,q_Aa,q_var,het_ratio;
int nliab,male_nliab;
double3 *pen;
double2 *male_pen;
char name[5],comment[81];
int nalleles;
struct allele_t *allele;
int chromosome;
void clear();
void create(int ma,int mp);
public:
int which_allele(char c);
int from_prepfile(data_file& file,int sexlinked);
int to_datafile(FILE *fp,int sex_linked=0,int strip_data=0);
virtual int input(data_file& file);
int real_input(data_file& file);
virtual int output(data_file& file);
int revise(editable_locus &);
locus& operator=(const locus &old);
locus& shrink_copy(const locus &old);
locus(int ma,int mp=MAXLIAB);
locus();
~locus();
char *label() { return name; }
char rand_allele();
};

class dcwindow;
class ls4parms;

class linksys_pedigree : public pedigree
{
protected:
char loops_broken,users_proband;
char fam_name[9];
linksys_person *proband;
data_object *new_own_type();
//person *new_own_type();
int numloci;
// locus loci[LOCPHENDIM];
locus *loci;
char have_condensed[LOCPHENDIM];
void set_none_condensed();
public:
are_loops_broken() { return loops_broken; }
char *locus_name(int i) { return loci[i].name; }
void show_loci(dcwindow* w=0);
join_loops();
break_loops();
do_pointers();
output_ped(int famnum,data_file &file,int loc_to_do,char locname[][5],
  ls4parms &lp);
output_bat(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkage="",
   int use_mlink=0);
output_script(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkage="",
   int use_mlink=0,int sex_linked=0);
output_lsp_script(char *fn,int loc_to_do,char locname[][5],float rf[],float sexratio,char *linkage="",
   int use_mlink=0,int sex_linked=0);
output_par(char *fn,int loc_to_do,char locname[][5],ls4parms &lp);
output_mflink_par(char *fn,int loc_to_do,char locname[][5],float rf[],ls4parms &lp);
#if 0
output_bat_or_sh(char *fn,int loc_to_do,char locname[][5],float rf[],
   float sexratio,char *linkage,int use_mlink,int use_ilink,int use_unix,
   float mlink_rf[],int mlink_rf_todo,int linkmap_evals,int avoid_lsp);
#else
output_bat_or_sh(char *fn,int loc_to_do,char locname[][5],float rf[],
   int sex_linked,ls4parms &l);
#endif
int sort_markers(char *fn,locus **l);
void set_loci(locus *l) { loci=l; }
int get_numloci() { return numloci; }
void set_numloci(int n) { numloci=n; }
linksys_pedigree(char *fn="",int dim=20);
randomise(char locname[][5],float rf);
set_pen(char locname[][5],float pen[3]);
output(data_file &file);
output_kin_par(data_file &file,int loc_to_do,char locname[][5]);
// fix(int l);
condense(int l,int towhat,data_file *fkey=0);
check(char *locname);
linksys_person *member(int i) { return (linksys_person *)pedigree::member(i); }
};

class editable_ls4parms;

#pragma ZTC align 2
class ls4parms : public data_object {
public:
char do_makeped,do_ppd,target_unix,use_mlink,use_crimap,
use_kin,use_lsp,mlink_diff_ratio,avoid_lsp,condkey,use_mflink,
strip_labels,use_vitesse,estimate_freqs;
short use_ilink,mlink_rf_todo,linkmap_evals;
float mlink_rf[20];
char loc_file[65];
dcstring input_file_name;
char linkage_path[65];
double sexratio;
char keep_fname;
load(char *fn,char *ver);
save(char *fn,char *ver);
ls4parms();
ls4parms& operator=(const ls4parms&);
revise(editable_ls4parms& lp_ed);
};
#pragma ZTC align

extern char *NL_STR;

#endif

