 /* PEDLS4.CPP */ 
/* Copyright Dave Curtis 1991 */ 
#include "pedlink.hpp"
#include "dcwindow.hpp"
#include "dcerror.hpp"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "hashfile.hpp"
#define USEDCERROR
#include "linkfile.h"
#ifdef MDB
#include <mem.h>
#endif

#ifndef max
#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))
#endif

void linksys_pedigree::set_none_condensed()
{
for (int i=0;i<LOCPHENDIM;++i)
  have_condensed[i]=0;
}

locphen_t::locphen_t() 
{ 
*loc=*phen='\0'; 
}

ppd_person::ppd_person()
{
probnum=psnum=msnum=offnum=0;
}

linksys_pedigree::linksys_pedigree(char *fn,int dim) : pedigree(dim)
{ 
pointers_done=users_proband=loops_broken=0; proband=0; strcpy(fam_name,fn); 
numloci=0;
set_none_condensed();
}

ped_person &ped_person::operator=(const ped_person &old)
{
person::operator=(old);
numloci=old.numloci;
famnum=old.famnum;
idnum=old.idnum;
fidnum=old.fidnum;
midnum=old.midnum;
for (int i=0;i<numloci;++i) 
  {
  bin[i].value=old.bin[i].value;
  bin[i].type=old.bin[i].type;
  }
return *this;
}

ped_person::ped_person()
{
numloci=idnum=fidnum=midnum=0;
for (int i=0;i<MAXLOCI;++i) bin[i].type=MARKER;
}

int linksys_pedigree::do_pointers()
{
if (pointers_done) return 1;
pedigree::do_pointers();
int i;
for (i=0;i<count;++i)
  member(i)->ped_num=i+1;
return (!dcerror.stat());
}

linksys_person::~linksys_person()
{
if (locphen) delete [] locphen;
}

int linksys_person::alloc_loci(int nl)
{
if (locphen) { delete [] locphen; locphen=NULL; }
nloci=0;
#if 0
if (nl && (locphen=new locphen_t[nl])==NULL) return dcerror(ENOMEM);
#else
if (nl!=0)
  {
  locphen=new locphen_t[nl];
  if (!locphen) return dcerror(ENOMEM);
  }
#endif
return nloci=nl;
}

linksys_person::linksys_person()
{
lsid[0]=fname[0]=group[0]=names[0]=age[0]=sex[0]=lsfid[0]=lsmid[0]='\0';
locphen=NULL;
nloci=0;
}

linksys_person &linksys_person::operator=(const linksys_person &old)
{
person::operator=(old);
ped_num=old.ped_num;
strcpy(fname,old.fname);
strcpy(lsid,old.lsid);
strcpy(group,old.group);
strcpy(names,old.names);
strcpy(age,old.age);
strcpy(sex,old.sex);
strcpy(lsfid,old.lsfid);
strcpy(lsmid,old.lsmid);
alloc_loci(old.nloci);
for (int i=0;i<nloci;++i) locphen[i]=old.locphen[i];
return *this;
}

int ppd_person::input(data_file &file)
{
char line[501],rest[301],buffer[21];
int c;
do 
  {
  char fname[20],buff[20]; 
  if (!fgets(line,500,get_fp(file)))
   return 0;
  c=sscanf(line," %19s %d %d %d %d %d %d %c %d %100[^\n]",
      buff,&idnum,&fidnum,&midnum,&offnum,&psnum,&msnum,&gender,&probnum,rest);
  if (atoi(buff)) sprintf(fname,"%03d",atoi(buff));
  else strcpy(fname,buff);
  if (c>=1 && !famnum.assign(fname)) { dcerror(ENOMEM); return 0; }
  } while (c<1); // ignore blank lines
if (c<9)
      { dcerror(2,"Bad PPD file format"); return 0; }
gender=(gender=='1')?'M':'F';
sprintf(buffer,"%s%03d",(char*)famnum,idnum);
if (!id.assign(buffer))
  dcerror(ENOMEM,"Out of memory in ppd_person::input()");
if (fidnum)
  {
  sprintf(buffer,"%s%03d",(char*)famnum,fidnum);
  if (!fid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
  }
else fid.strfree();
if (midnum)
  {
  sprintf(buffer,"%s%03d",(char*)famnum,midnum);
  if (!mid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
  }
else mid.strfree();
fill_markers(rest);
return 1;
}

int ped_person::input(data_file &file)
{
char line[501],rest[301],buffer[21];
int c;
do 
  { 
  char fname[20],buff[20];
  if (!fgets(line,500,get_fp(file)))
   return 0;
  c=sscanf(line," %19s %d %d %d %c %100[^\n]",
      buff,&idnum,&fidnum,&midnum,&gender,rest);
  if (atoi(buff)) sprintf(fname,"%03d",atoi(buff));
  else strcpy(fname,buff);
  if (c>=1 && !famnum.assign(fname)) { dcerror(ENOMEM); return 0; }
  } while (c<1); // ignore blank lines
if (c<5)
      { dcerror(2,"Bad PED file format"); return 0; }
gender=(gender=='1')?'M':'F';
sprintf(buffer,"%s%03d",(char*)famnum,idnum);
if (!id.assign(buffer))
  dcerror(ENOMEM,"Out of memory in ped_person::input()");
if (fidnum)
  {
  sprintf(buffer,"%s%03d",(char*)famnum,fidnum);
  if (!fid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
  }
else fid.strfree();
if (midnum)
  {
  sprintf(buffer,"%s%03d",(char*)famnum,midnum);
  if (!mid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
  }
else mid.strfree();
fill_markers(rest);
return 1;
}

int ped_person::fill_markers(char *rptr) // from datafile
{
char buff[30];
char *bptr=buff;
numloci=0;
do 
  {
  if (isspace(*rptr))
     {
     ++rptr;
     if (isspace(*rptr))
       {
       *bptr='\0';
       bin[numloci++].value=buff;
       bptr=buff;
       while (*rptr && isspace(*rptr)) ++rptr;
       }
     }
  else
    *bptr++=*rptr++;
  } while (*rptr);
*bptr='\0';
if (strlen(buff))
   bin[numloci++].value=buff;
return 1;
}

int linksys_person::input(data_file &file)
{
char line[2000],buffer[9];
int len,c;
if (!fgets(line,2000,get_fp(file)))
      return 0;
if (strchr(line,'\n')) *strchr(line,'\n')='\0';
*fname='\0';
c=sscanf(line,"%8[^\n]%4[^\n]%1[^\n]%48[^\n]%6[^\n]%1[^\n]%4[^\n]%4[^\n]%n",
  fname,
  lsid,
  group,
  names,
  age,
  sex,
  lsfid,
  lsmid,&len);
if (c!=8) return dcerror(2,"Line too short in SDF file: %s",line);
int i=((int)strlen(line)-len-1)/12+1; // strlen returns size_t - unsigned
if (!alloc_loci(i) && i) return 0; // ENOMEM, or if OK nloci is set
for (i=0;i<nloci;++i)
  {
  sscanf(line+len+i*12,"%4[^\n]%8[^\n]",
     locphen[i].loc,locphen[i].phen);
  char *ptr;
  if ((ptr=strchr(locphen[i].loc,' '))!=0) *ptr='\0';
  for (int j=strlen(locphen[i].phen)-1;j>=0 && locphen[i].phen[j]==' ';--j)
     locphen[i].phen[j]='\0';  // clear trailing blanks only
//  if ((ptr=strchr(locphen[i].phen,' '))!=0) *ptr='\0';
  }
gender=toupper(sex[0]);
if (gender!='M' && gender !='F')
  return dcerror(2,"Bad gender code in SDF file: %s",line);
if (sscanf(lsid," %s ",buffer)!=1)
  return dcerror(2,"No subject ID in SDF file: %s",line);
strcpy(lsid,buffer);
if (sscanf(fname," %s ",buffer)!=1)
  return dcerror(2,"No pedigree ID in SDF file: %s",line);
strcpy(fname,buffer);
sprintf(line,"%s%s",fname,lsid);
if (!id.assign(line))
   return dcerror(ENOMEM,"Out of memory in linksys_person::input()");
*buffer='\0';
sscanf(lsfid," %s ",buffer);
strcpy(lsfid,buffer);
*buffer='\0';
sscanf(lsmid," %s ",buffer);
strcpy(lsmid,buffer);
if (*lsfid)
  {
  sprintf(line,"%s%s",fname,lsfid);
  if (!fid.assign(line)) return dcerror(ENOMEM,"Out of memory in linksys_person::input()");
  }
if (*lsmid)
  {
  sprintf(line,"%s%s",fname,lsmid);
  if (!mid.assign(line)) return dcerror(ENOMEM,"Out of memory in linksys_person::input()");
  }
return 1;
}

data_object *linksys_pedigree::new_own_type()
{ return new linksys_person; }

ped_person &ppd_person::from_linksys(const linksys_person& old)
{
ped_person::from_linksys(old);
probnum=atoi(old.group);
person *par;
int ch,sp;
if ((par=old.get_father())!=0)
     {
     for (sp=0;sp<par->numspo;++sp)
       for (ch=0;ch<par->spouse[sp]->get_numch();++ch)
         if (par->spouse[sp]->get_child(ch)==(person*)&old) goto found1;
       dcerror(2,"Could not match child with father (%s with %s)",
         (char*)id,(char*)par->id);
       return *(ped_person*)this; // should never happen
     found1:
     if (ch+1 < par->spouse[sp]->get_numch())
        psnum=((linksys_person*)par->spouse[sp]->get_child(ch+1))->ped_num;
     else if (sp+1<par->get_numspo() && par->spouse[sp+1]->get_numch())
        psnum=((linksys_person*)par->spouse[sp+1]->get_child(0))->ped_num;
     }
if ((par=old.get_mother())!=0)
     {
     for (sp=0;sp<par->numspo;++sp)
       for (ch=0;ch<par->spouse[sp]->get_numch();++ch)
         if (par->spouse[sp]->get_child(ch)==(person*)&old) goto found2;
       dcerror(2,"Could not match child with mother (%s with %s)",
         (char*)id,(char*)par->id);
       return *(ped_person*)this;
     found2:
     if (ch+1 < par->spouse[sp]->get_numch())
        msnum=((linksys_person*)par->spouse[sp]->get_child(ch+1))->ped_num;
     else if (sp+1<par->numspo && par->spouse[sp+1]->get_numch())
        msnum=((linksys_person*)par->spouse[sp+1]->get_child(0))->ped_num;
     }
if (old.numspo && old.spouse[0]->get_numch())
    offnum=((linksys_person*)old.spouse[0]->get_child(0))->ped_num;
return *(ped_person*)this;
}

ped_person &ped_person::from_linksys(const linksys_person& old)
{
if (!id.assign(old.id)) dcerror(ENOMEM,"Out of memory in ped_person::from_linksys()");
parents=old.parents;
gender=old.gender;
idnum=old.ped_num;
famnum=old.get_fname(); // maybe overwritten later by output_ped()
if (old.get_father())
     fidnum=((linksys_person*)old.get_father())->ped_num;
if (old.get_mother())
     midnum=((linksys_person*)old.get_mother())->ped_num;
return *this;
}

int ppd_person::output(data_file &file,int strip_labels)
{
fprintf(get_fp(file),"%s  %3d %3d %3d %3d %3d %3d %2d %2d  ",
        (char*)famnum,
        idnum,
        fidnum,
        midnum,
        offnum,
        psnum,
        msnum,
        gender=='M'?1:2,
        probnum);
for (int i=0;i<numloci;++i)
   {
   if (bin[i].type==QUAN || bin[i].type==MARKER)
      fprintf(get_fp(file),"%s ",(char*)bin[i].value);
   else if (bin[i].type==AFF)
      fprintf(get_fp(file),"%c %s ",
         (char)bin[i].value[0],(char*)bin[i].value+1);
   else for (int j=0;j<strlen(bin[i].value);++j)
     fprintf(get_fp(file),"%c ",(char)bin[i].value[j]);
   fprintf(get_fp(file),"%s",(i!=numloci-1)?" ":"");
   }
if (strip_labels) fprintf(get_fp(file)," %s",NL_STR);
else fprintf(get_fp(file),"  <<- %s%s",(char*)id,NL_STR);
return 1;
}

int ped_person::output(data_file &file,int strip_labels)
{
#if 0
if (dokey && !*keyfn)
  {
  strncpy(keyfn,filename,60);
  keyfn[60]='\0';
  char *ptr;
  if ((ptr=strchr(keyfn,'.'))!=0) strcpy(ptr,".KEY");
  else strcat(keyfn,".KEY");
  keyfile=fopen(keyfn,"w");
  }
keyfile && fprintf(keyfile,"%s  %3d  <- %s\n",(char*)famnum,idnum,(char*)id);
#endif
fprintf(get_fp(file),"%s  %3d %3d %3d %2d  ",
        (char*)famnum,
        idnum,
        fidnum,
        midnum,
        gender=='M'?1:2);
for (int i=0;i<numloci;++i)
   {
   if (bin[i].type==QUAN || bin[i].type==MARKER)
      fprintf(get_fp(file),"%s ",(char*)bin[i].value);
   else if (bin[i].type==AFF)
      fprintf(get_fp(file),"%c %s ",
         (char)bin[i].value[0],(char*)bin[i].value+1);
   else for (int j=0;j<strlen(bin[i].value);++j)
     fprintf(get_fp(file),"%c ",(char)bin[i].value[j]);
   fprintf(get_fp(file),"%s",(i!=numloci-1)?" ":"");
   }
if (strip_labels) fprintf(get_fp(file)," %s",NL_STR);
else fprintf(get_fp(file),"  <<- %s%s",(char*)id,NL_STR);
// new! I think this should be OK though
return 1;
}

int ppd_person::input(data_file &file,int n,dc_locus_def_t *def)
{
char line[501],rest[401],buffer[21];
int c;
do 
  {
  char fname[20],buff[20]; 
  if (!fgets(line,500,get_fp(file)))
   return 0;
  c=sscanf(line," %19s %d %d %d %d %d %d %c %d %400[^\n]",
      buff,&idnum,&fidnum,&midnum,&offnum,&psnum,&msnum,&gender,&probnum,rest);
  if (atoi(buff)) sprintf(fname,"%03d",atoi(buff));
  else strcpy(fname,buff);
  if (c>=1 && !famnum.assign(fname)) { dcerror(ENOMEM); return 0; }
  } while (c<1); // ignore blank lines
if (c<9)
      { dcerror(2,"Bad PPD file format"); return 0; }
gender=(gender=='1')?'M':'F';
sprintf(buffer,"%s%03d",(char*)famnum,idnum);
if (!id.assign(buffer))
  dcerror(ENOMEM,"Out of memory in ppd_person::input()");
if (fidnum)
  sprintf(buffer,"%s%03d",(char*)famnum,fidnum);
else *buffer='\0';
if (!fid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
if (midnum)
  sprintf(buffer,"%s%03d",(char*)famnum,midnum);
else *buffer='\0';
if (!mid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ppd_person::input()");
return fill_markers(rest,n,def);
}

int ped_person::input(data_file &file,int n,dc_locus_def_t *def)
{
char line[501],rest[401],buffer[21];
int c;
do 
  { 
  char fname[20],buff[20];
  if (!fgets(line,500,get_fp(file)))
   return 0;
  c=sscanf(line," %19s %d %d %d %c %400[^\n]",
      buff,&idnum,&fidnum,&midnum,&gender,rest);
  if (atoi(buff)) sprintf(fname,"%03d",atoi(buff));
  else strcpy(fname,buff);
  if (c>=1 && !famnum.assign(fname)) { dcerror(ENOMEM); return 0; }
  } while (c<1); // ignore blank lines
if (c<5)
      { dcerror(2,"Bad PED file format"); return 0; }
gender=(gender=='1')?'M':'F';
sprintf(buffer,"%s%03d",(char*)famnum,idnum);
if (!id.assign(buffer))
  dcerror(ENOMEM,"Out of memory in ped_person::input()");
if (fidnum)
  sprintf(buffer,"%s%03d",(char*)famnum,fidnum);
else *buffer='\0';
if (!fid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ped_person::input()");
if (midnum)
  sprintf(buffer,"%s%03d",(char*)famnum,midnum);
else *buffer='\0';
if (!mid.assign(buffer))
     dcerror(ENOMEM,"Out of memory in ped_person::input()");
return fill_markers(rest,n,def);
}

int locus_string_input(char *str,dc_locus_def_t *def,char **line)
{
int nread,n,i,a1,a2;
char buff[30];
switch (def->type)
  {
  case DCL_AFF:
    if (def->info.aff.nliab==1)
      {
      if (sscanf(*line," %s%n",str,&nread)!=1) return 0;
      }
    else
      if (sscanf(*line," %s %s%n",str,buff,&nread)!=2) return 0;
      else strcat(str,buff);
    break;
  case DCL_QUAN:
    if (sscanf(*line," %s%n",str,&nread)!=1) return 0;
    break;
  case DCL_BINARY:
    for (i=0,nread=0;i<def->info.bin.nfact;++i,nread+=n)
      if (sscanf(*line," %c%n",&str[i],&n)!=1) return 0;
    str[i]='\0';
    break;
  case DCL_NUMBERED:
    if (sscanf(*line," %d %d%n",&a1,&a2,&nread)!=2) return 0;
    sprintf(str,"%c%c",a1<10?'0'+a1:'a'+a1-10,a2<10?'0'+a2:'a'+a2-10);
    break;
  }    
*line+=nread;
return 1;
}

int ped_person::fill_markers(char *rptr,int n,dc_locus_def_t *def)
{
char buff[30],*saved;
saved=rptr;
for (numloci=0;numloci<n;++numloci)
  {
  if (!locus_string_input(buff,&def[numloci],&rptr))
    { dcerror(2,"Bad pedigree file data:  %s",saved); return 0; }
  else
    bin[numloci].value=buff;
  }
return 1;
}



