/* PEDTABLE.CPP */
/* Copyright Dave Curtis 1991 */
#include "pedtable.hpp"
#include "pedlink.hpp"
#include <dcerror.hpp>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#ifdef MDB
#include <mem.h>
#endif

#ifndef max
#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))
#endif

//void *tab_person::base() { return this; }

tab_person& tab_person::operator=(const tab_person& old)
{
person::operator=(old);
posy=old.posy;
gennum=old.gennum;
if (inset_str) { free(inset_str); inset_str=0; }
if (old.inset_str && (inset_str=strdup(old.inset_str))==0)
  dcerror(1,"Out of memory in tab_person::operator=()");
for (int i=0;i<ROWS;++i)
  {
  if (lines[i]) { free(lines[i]); lines[i]=0; }
  if (old.lines[i] && (lines[i]=strdup(old.lines[i]))==0)
     dcerror(1,"Out of memory in tab_person::operator=()");
  }
return *this;
}

tab_person& tab_person::from_ls(const linksys_person& old)
{
int i;
person::operator=(old);
for (i=0;i<ROWS;++i)
  if (lines[i]) { free(lines[i]); lines[i]=0; }
for (i=0;i<min(ROWS,LOCPHENDIM);++i)
  if (i<old.nloci && *old.locphen[i].phen  && (lines[i]=strdup(old.locphen[i].phen))==0)
     dcerror(1,"Out of memory in tab_person::from_ls()");
if (lines[0])
   {
   char diag_code;
   diag_code=strchr(lines[0],'-')?*(strchr(lines[0],'-')+1):lines[0][0];
   if (diag_code>'1' && diag_code<='9')
      diag=diag_code-'1';
   else diag=0;
   }
return *this;
}

tab_person& tab_person::from_ped(const ped_person& old)
{
int i;
person::operator=(old);
for (i=0;i<ROWS;++i)
  if (lines[i]) { free(lines[i]); lines[i]=0; }
for (i=0;i<min(ROWS,old.numloci);++i)
  if ((lines[i]=strdup(old.bin[i].value))==0)
     dcerror(ENOMEM,"Out of memory in tab_person::from_ls()");
if (lines[0])
   diag=max(lines[0][0]-'1',0);
return *this;
}

void tab_person::clear()
{
draw_stat=NOTDONE; gennum=0; posy=-1;
if (gender=='M') for (int i=0;i<numspo;++i) spouse[i]->set_stat(NOTDONE);
}

table_item::table_item(float dropping)
{ draw_stat=NOTDONE; ver_drop=dropping; }

symbol_above_child::symbol_above_child() : symbol(0.5)
{ ch=''; str="      "; }

void symbol_above_child::draw(gr_table_output_device *dev)
{ dev->draw(0.5,0.0,0.5,1); }

symbol_first_marriage_lo_half::symbol_first_marriage_lo_half(int fertile)
 : symbol(fertile?0.5:-1)
{ ch=fertile?'':''; str=fertile?"    ":"  Ĵ   ";}

void symbol_first_marriage_lo_half::draw(gr_table_output_device *dev)
{
dev->draw(0,0.5,1,0.5);
dev->draw(0.3,0.5,0.3,1);
if (ch=='') dev->draw(0.5,0.5,0.5,0);
}

symbol_first_marriage_hi_half::symbol_first_marriage_hi_half()
 : symbol(0.3)
{ch='';str="      ";}

void symbol_first_marriage_hi_half::draw(gr_table_output_device *dev)
{ dev->draw(0.3,0.5,1,0.5); dev->draw(0.3,0.5,0.3,0); }

symbol_middle_marriage_lo_half::symbol_middle_marriage_lo_half(int fertile)
 : symbol(fertile?0.5:-1)
{ ch=fertile?'':''; str=fertile?"     ":"      "; }

void symbol_middle_marriage_lo_half::draw(gr_table_output_device *dev)
{
dev->draw(0.3,0.5,1,0.5);
dev->draw(0.3,0.5,0.3,1);
if (ch=='') dev->draw(0.5,0.5,0.5,0);
}

symbol_middle_marriage_hi_half::symbol_middle_marriage_hi_half()
 : symbol(0.3)
{ch='';str="     ";}

void symbol_middle_marriage_hi_half::draw(gr_table_output_device *dev)
{
dev->draw(0,0.5,1,0.5);
dev->draw(0.3,0.5,0.3,0);
}

symbol_last_marriage_lo_half::symbol_last_marriage_lo_half(int fertile)
: symbol(fertile?0.5:-1)
{ ch=fertile?'':''; str=fertile?"     ":"      ";}

void symbol_last_marriage_lo_half::draw(gr_table_output_device *dev)
{ dev->draw(0.3,0.5,1,0.5); dev->draw(0.3,0.5,0.3,1);
if (ch=='') dev->draw(0.5,0.5,0.5,0); }

symbol_last_marriage_hi_half::symbol_last_marriage_hi_half()
 : symbol(0.3)
{ch='';str="      ";}

void symbol_last_marriage_hi_half::draw(gr_table_output_device *dev)
{
dev->draw(0.3,0.5,0.3,0);
dev->draw(0,0.5,0.3,0.5);
}

symbol_only_marriage::symbol_only_marriage(int fertile)
: symbol(0.5)
{ ch=fertile?'':''; str=fertile?"     ":"      "; }

void symbol_only_marriage::draw(gr_table_output_device *dev)
{ dev->draw(0,0.5,1,0.5);
if (ch=='') dev->draw(0.5,0.5,0.5,0); }

symbol_many_children::symbol_many_children()
{ ch='';str="  Ĵ   "; }

void symbol_many_children::draw(gr_table_output_device *dev)
{ dev->draw(0,0.5,1,0.5); dev->draw(0.5,0.5,0.5,1); }

symbol_only_child::symbol_only_child()
{ ch='';str="   Ŀ   "; }

void symbol_only_child::draw(gr_table_output_device *dev)
{ //dev->draw(0.5,0,0.5,0.5,0);
dev->draw(0,0.5,0.5,0.5); dev->draw(0.5,0.5,0.5,1); }

symbol_between_spouses_lo_half::symbol_between_spouses_lo_half()
 : symbol(0)
{ ch='';str="      "; }

void symbol_between_spouses_lo_half::draw(gr_table_output_device *dev)
{ dev->draw_loopable_line(); }

symbol_between_spouses_hi_half::symbol_between_spouses_hi_half()
 : symbol(0)
{ ch='';str="      "; }

void symbol_between_spouses_hi_half::draw(gr_table_output_device *dev)
{ dev->draw_loopable_line(); }

symbol_between_sibs::symbol_between_sibs()
 : symbol(0)
{ ch='';str="      "; }

void symbol_between_sibs::draw(gr_table_output_device *dev)
{ dev->draw_loopable_line(); }

symbol_first_sib::symbol_first_sib()
 : symbol(0.5)
{ ch='';str="      "; }

void symbol_first_sib::draw(gr_table_output_device *dev)
{ dev->draw(0.5,0.5,1,0.5); dev->draw(0.5,0.5,0.5,0); }

symbol_mid_sib::symbol_mid_sib()
 : symbol(0.5)
{ ch='';str="     "; }

void symbol_mid_sib::draw(gr_table_output_device *dev)
{ dev->draw(0,0.5,1,0.5); dev->draw(0.5,0.5,0.5,0); }

symbol_last_sib::symbol_last_sib()
 : symbol(0.5)
{ ch='';str="      "; }

void symbol_last_sib::draw(gr_table_output_device *dev)
{ dev->draw(0,0.5,0.5,0.5); dev->draw(0.5,0.5,0.5,0); }

void person_ptr::print(char *s,int line,int length,int midx,int midy)
{ ptr->print(s,line,length); }


table::table(int h,int w)
{
max_h=h*2;
max_w=w*2;
item=new TABLE_ROW[max_h];
extra_line=new ROW_LIST[max_h]; // array of empty lists
if (!item || !extra_line) { dcerror(1,"Not enough memory for table"); return; }
for (int i=0;i<max_h;++i)
  {
//   extra_line[i].setflush(1); no point really
  item[i]=new PTABLE_ITEM[max_w];
  if (!item[i]) { dcerror(1,"Not enough memory for table"); return; }
  }
for (i=0;i<max_h;++i)
  for (int j=0;j<max_w;++j)
    item[i][j]=0;
max_posy=hi_gen=0;
ntl=0;
for (i=0;i<ROWS+1;++i) t_lines[i]=0;
}

void table::clear()
{
int i,j;
ancs_to_do.cleanup();
for (i=0;i<max_h;++i)
  {
  for (j=0;j<max_w;++j)
    if (item[i][j])
      {
      item[i][j]->clear();
      delete item[i][j];
      item[i][j]=0;
      }
  while (extra_line[i].size())
    {
    TABLE_ROW oldrow=extra_line[i].linkout();
    for (j=0;j<max_w;++j)
      if (oldrow[j]) delete oldrow[j];
    delete oldrow;
    }
  }
max_posy=hi_gen=0;
}

table::~table()
{
clear();
for (int i=0;i<max_h;++i)
  delete item[i];
delete [] item;
delete [] extra_line;
}

void person_ptr::draw(gr_table_output_device *dev)
{
if (dev->unsex || ptr->gender=='X')
   dev->draw_diamond(dev->no_diag?(ptr->get_diag()/100)*100:ptr->get_diag(),
     dev->inset?ptr->get_inset_str():0);
else if (ptr->gender=='M')
   dev->draw_square(dev->no_diag?(ptr->get_diag()/100)*100:ptr->get_diag(),
     dev->inset?ptr->get_inset_str():0);
else
   dev->draw_circle(dev->no_diag?(ptr->get_diag()/100)*100:ptr->get_diag(),
     dev->inset?ptr->get_inset_str():0);
}

tab_person::tab_person()
{
inset_str=0;
for (int i=0;i<ROWS;++i) lines[i]=0;
clear();
}

tab_person::~tab_person()
{
for (int i=0;i<ROWS;++i) if (lines[i]) free(lines[i]);
}

void tab_pedigree::clear_draw_data()
{
for (int i=0;i<count;++i)
   member(i)->clear();
}

int tab_pedigree::save(data_file &file)
{
for (int i=0;i<count;++i)
   if (!member(i)->output(file)) return 0;
return 1;
}

int tree_data_file::output(tab_person &temp)
{
fprintf("%s\n%c,%d",temp.id.is_null()?"":(char*)temp.id,temp.gender,temp.diag);
if (temp.inset_str)
  fprintf(" \"%s\"",temp.inset_str);
fprintf("\n%s,%s\n",temp.fid.is_null()?"":(char*)temp.fid,
  temp.mid.is_null()?"":(char*)temp.mid);
for (int i=0;i<ROWS;++i)
  {
  fprintf("%s\n",temp.lines[i]?temp.lines[i]:"");
  }
fprintf("\n");
return (ferror(fp))?
  dcerror(3,"Error writing file"),0:1;
}

int tab_person::output(data_file &file)
{
fprintf(get_fp(file),"%s\n%c,%d",id.is_null()?"":(char*)id,gender,diag);
if (inset_str) fprintf(get_fp(file),"\"%s\"",inset_str);
fprintf(get_fp(file),"\n%s,%s\n",fid.is_null()?"":(char*)fid,
       mid.is_null()?"":(char*)mid);
for (int i=0;i<ROWS;++i)
  {
  fprintf(get_fp(file),"%s\n",lines[i]?lines[i]:"");
  }
fprintf(get_fp(file),"\n");
return (ferror(get_fp(file)))?
  dcerror(3,"Error writing file"),0:1;
}

int tab_person::input(data_file &file)
{
char line[241],buff1[MAXSTR+1],buff2[MAXSTR+1];
int c;
do    {					 // read in initial blank or bad lines
      *buff1='\0';
      if (feof(get_fp(file)) || !fgets(line,240,get_fp(file)))
         return 0;
      if (sscanf(line," %20s",buff1)!=1) continue;
      if (!strcmp(buff1,"END") || !strcmp(buff1,"end")) return 0;
      } while (!isalnum(*buff1) && !isprint(*buff1));
if (!id.assign(buff1))
  return dcerror(ENOMEM,"Out of memory in tab_person::input()");
if (!fgets(line,240,get_fp(file))) return 1;
*buff2=*buff1='\0';
sscanf(line," %c , %d \"%[^\n\"]",&gender,&diag,buff1);
if (*buff1)
  {
  if ((inset_str=strdup(buff1))==0)
    return dcerror(ENOMEM,"Out of memory in tab_person::input()");
  *buff1='\0';
  }
gender=toupper(gender);
if (gender!='M'&&gender!='F'&&gender!='X')
    dcerror(23,"Bad gender : %c in %s",gender,(char*)id);
if (!fgets(line,240,get_fp(file))) return 1;
c=sscanf(line," %20[^ ,] , %20s",buff1,buff2);
if (c==1)
    dcerror(22,"Father but no mother in %s",(char*)id);
else if (c==2)
  {
  if (!fid.assign(buff1) || !mid.assign(buff2))
     return dcerror(ENOMEM,"Out of memory in tab_person::input()");
  }
for (int l=0;l<ROWS;++l)
     {
     if (!fgets(line,240,get_fp(file))) break;
     if (1==sscanf(line,"%20[^\n]",buff1))
         if ((lines[l]=strdup(buff1))==0)
           return dcerror(ENOMEM,"Out of memory in tab_person::input()");
     }
return 1;
}

int tree_data_file::input(tab_person &temp)
{
char line[241],buff1[MAXSTR+1],buff2[MAXSTR+1];
int c;
do    {					 // read in initial blank or bad lines
      *buff1='\0';
      if (!fgets(line,240))
         return 0;
      if (sscanf(line," %20s",buff1)!=1) continue;
      if (!strcmp(buff1,"END") || !strcmp(buff1,"end")) return 0;
      } while (!isalnum(*buff1) && !isprint(*buff1));
if (!temp.id.assign(buff1))
   return dcerror(ENOMEM,"Out of memory in tree_data_file::input()");
if (!fgets(line,240)) return 1;
*buff2=*buff1='\0';
sscanf(line," %c , %d \"%[^\n\"]",&temp.gender,&temp.diag,buff1);
if (*buff1)
  {
  if ((temp.inset_str=strdup(buff1))==0)
    return dcerror(ENOMEM,"Out of memory in tree_data_file::input()");
  *buff1='\0';
  }
temp.gender=toupper(temp.gender);
if (temp.gender!='M'&&temp.gender!='F'&&temp.gender!='X')
    dcerror(23,"Bad gender : %c in %s",temp.gender,(char*)temp.id);
if (!fgets(line,240)) return 1;
c=sscanf(line," %20[^ ,] , %20s",buff1,buff2);
if (c==1)
    dcerror(22,"Father but no mother in %s",(char*)temp.id);
else if (c==2)
  {
  if (!temp.fid.assign(buff1) || !temp.mid.assign(buff2))
     return dcerror(ENOMEM,"Out of memory in tree_data_file::input()");
  }
for (int l=0;l<ROWS;++l)
     {
     if (!fgets(line,240)) break;
     if (1==sscanf(line,"%20[^\n]",buff1))
         if ((temp.lines[l]=strdup(buff1))==0)
            return dcerror(ENOMEM,"Out of memory in tree_data_file::input()");
     }
return 1;
}

void table::set_t_line(int l,int offon)
{
int i;
t_lines[l]=offon;
for (ntl=0,i=0;i<ROWS+1;++i)
  if (t_lines[i]) ++ntl;
}

int table::get_t_line(int l)
{
return t_lines[l];
}

void table::print(char *s,int r,int c,int l)
{
if (item[r][c])
   {
   if (!(r%2))
     {
     for (int i=0,which_line=0;i<l;++which_line)
         if (t_lines[which_line]) ++i;
     item[r][c]->print(s,which_line,MAXSTR,MIDSTR,!(c%2)?MSCOL:MGAPCOL);
     }
   else
     item[r][c]->print(s,l,GAPROW,MGAPROW,!(c%2)?MSCOL:MGAPCOL);
   }
else
  {
  if (!(r%2))
     sprintf(s,"%*c",MAXSTR,' ');
  else
     sprintf(s,"%*c",GAPROW,' ');
  }
}

int table::ntlines(int r,int c)
{
return (c%2)?GAPCOL:ntl;
}

int table::nglines(int r,int c)
{
return ntl;
}

void symbol::print(char*s,int line,int length,int midx,int midy)
{
int start,j;
start=line<midy?0:line==midy?3:6;
for (j=0;j<midx && j<length;++j) s[j]=str[start];
if (j<length) s[j++]=str[start+1];
for (;j<length;++j) s[j]=str[start+2];
s[j]='\0';
}

void tab_person::print()
{
char dummy[80];
print(dummy,0,79);
puts(dummy);
for (int i=1;i<=ROWS;++i)
   print(dummy,i,79),puts(dummy);
}

void tab_person::print(char*s,int l,int length)
{
sprintf(s,"%*c",abs(length),' ');
if (l==0)
  {
  if (length>0  || strlen(id)<abs(length))
    {
    strncpy(s,id,min(strlen(id),abs(length)));
    if (length > 0 && length>strlen(id)+1) s[strlen(id)+1]=gender;
    }
  else
    {
    length*=-1;
    char *ptr=(char*)id+strlen(id)-length;
    for (int i=0;i<length;++i)
      {
      if (!*ptr) break;
      else if (*ptr!='0' || strchr(id,'-'))
        {
        strncpy(s,ptr,length-i);
        break;
        }
      else if (i<length/2)
        *s++=' ';
      ++ptr;
      }
    }
  }
else
  if (lines[l-1])
    {
    if (length>0  || strlen(lines[l-1])<abs(length))
       strncpy(s,lines[l-1],min(strlen(lines[l-1]),abs(length)));
    else
       strncpy(s,lines[l-1]+strlen(lines[l-1])+length,length*-1);
    }
}

data_object *tab_pedigree::new_own_type()
{ return new tab_person; }


