/* PEDTABLE.HPP */
/* Copyright Dave Curtis 1991 */
#ifndef PEDTABLEHPP
#define PEDTABLEHPP 1

#include <dcdlist.hpp>
#include "pedout.hpp"
#include "pedigree.hpp"

#define MAXWID 200
#define MAXGEN 8
#define ROWS 10  /* same as in pedout.hpp */
#define MAXSTR 20
#define MIDSTR 8

enum draw_status {NOTDONE,SEEDDONE,DONE,REP};

class tree_data_file : public data_file
{
public:
tree_data_file() {;}
~tree_data_file() {;}
virtual int input(tab_person &temp);
virtual int output(tab_person &temp);
};

class editable_tab_person;
class editable_pedparms;

class tab_person : public person
{
friend class person_ptr;
friend class tree_data_file;
friend class table;
int posy,width,gennum;
tab_person *set_new_anc();
draw_status draw_stat;
protected:
char *lines[ROWS],*inset_str;
public:
void set_diag(int d,char g='M') { gender=g; diag=d; }
char *get_inset_str() { return inset_str; }
int revise(editable_tab_person &);
tab_person &operator=(const tab_person& old);
tab_person &from_ls(const linksys_person& old);
tab_person &from_ped(const ped_person& old);
void clear();
void set_draw_stat(draw_status ds) { draw_stat=ds; }
draw_status get_draw_stat() { return draw_stat; }
tab_person();
~tab_person();
void print(char *s,int line,int length=MAXSTR);
void print();
// void *base(); // { return this; }
virtual int input(data_file& file);
int ged_input(data_file& file);
virtual int output(data_file& file);
// #define tab_person_ptr(x) ((x)?(tab_person*)(x):0)
// #define tab_person_ptr(x) ((x)?((tab_person*)((x)->base())):0)
};

class table_item
{
friend class table;
draw_status draw_stat;
protected:
float ver_drop;
// -1 : stop a dropping line
// 0 : allow a dropping line to go through
// otherwise set a position value for dropping line
// - the only thing a dropping line should ever hit
// is a loopable line
public:
virtual void print(char *s,int line,int len,int midx,int midy)=0;
virtual void draw(gr_table_output_device *dev)=0;
virtual void clear() { draw_stat=NOTDONE; }
virtual int is_a_person() { return 0; }
table_item(float dropping=-1);
virtual ~table_item() {;}
draw_status get_draw_stat() { return draw_stat; }
float drops() { return ver_drop; }
};


class symbol : public table_item
{
protected:
char ch,*str;
// marriage_status marr_stat;
public:
void print(char *s,int line,int len,int midx,int midy);
virtual void draw(gr_table_output_device *dev)=0;
symbol(float dropping=-1) : table_item(dropping) {;}
~symbol() {;}
};

class symbol_above_child:public symbol
{
public:
symbol_above_child();
void draw(gr_table_output_device *dev);
};

class symbol_first_marriage_lo_half:public symbol
{
public:
symbol_first_marriage_lo_half(int fertile);
void draw(gr_table_output_device *dev);
};

class symbol_first_marriage_hi_half:public symbol
{
public:
symbol_first_marriage_hi_half();
void draw(gr_table_output_device *dev);
};

class symbol_middle_marriage_lo_half:public symbol
{
public:
symbol_middle_marriage_lo_half(int fertile);
void draw(gr_table_output_device *dev);
};

class symbol_middle_marriage_hi_half:public symbol
{
public:
symbol_middle_marriage_hi_half();
void draw(gr_table_output_device *dev);
};

class symbol_last_marriage_lo_half:public symbol
{
public:
symbol_last_marriage_lo_half(int fertile);
void draw(gr_table_output_device *dev);
};

class symbol_last_marriage_hi_half:public symbol
{
public:
symbol_last_marriage_hi_half();
void draw(gr_table_output_device *dev);
};

class symbol_only_marriage:public symbol
{
public:
symbol_only_marriage(int fertile);
void draw(gr_table_output_device *dev);
};

class symbol_many_children:public symbol
{
public:
symbol_many_children();
void draw(gr_table_output_device *dev);
};

class symbol_only_child:public symbol
{
public:
symbol_only_child();
void draw(gr_table_output_device *dev);
};

class symbol_between_spouses_lo_half:public symbol
{
public:
symbol_between_spouses_lo_half();
void draw(gr_table_output_device *dev);
};

class symbol_between_spouses_hi_half:public symbol
{
public:
symbol_between_spouses_hi_half();
void draw(gr_table_output_device *dev);
};

class symbol_between_sibs:public symbol
{
public:
symbol_between_sibs();
void draw(gr_table_output_device *dev);
};

class symbol_first_sib:public symbol
{
public:
symbol_first_sib();
void draw(gr_table_output_device *dev);
};

class symbol_mid_sib:public symbol
{
public:
symbol_mid_sib();
void draw(gr_table_output_device *dev);
};

class symbol_last_sib:public symbol
{
public:
symbol_last_sib();
void draw(gr_table_output_device *dev);
};

class person_ptr :public table_item
{
tab_person *ptr;
public:
int is_a_person() { return 1; }
void draw(gr_table_output_device *dev);
void print(char *s,int line,int length,int midx,int midy);
person_ptr(tab_person *p) {ptr=p;}
~person_ptr() {;}
void clear() { ptr->clear(); }
};

enum which_marriage {FIRST,MIDDLE,LAST,ONLY};

typedef table_item *PTABLE_ITEM;
typedef PTABLE_ITEM *TABLE_ROW;
typedef TABLE_ROW *PTABLE_ROW;
declare (dczGDList,PTABLE_ITEM);
typedef dczGDList(PTABLE_ITEM) ROW_LIST;

class table
{
protected:
int max_h,max_w;
int person_char_height,
    symbol_char_height,
    person_char_width,
    symbol_char_width;
int max_posy,hi_gen;
int extend;
int ntl;
char *title;
int t_lines[ROWS+1];
typedef tab_person *PTAB_PERSON;
declare (dczGDList,tab_person)
; // having this at end of previous line caused error
  // with Sun compiler, but maybe needed for others?
dczGDList(tab_person) ancs_to_do;
public:
ROW_LIST *extra_line;
TABLE_ROW *item;
char *get_title() { return title; }
void set_title(char *t) { title=t; }
add_extra_line(int l,int at_top=0);
void set_t_line(int l,int offon);
get_t_line(int l);
int get_max_posy() { return max_posy; }
int get_num_col_pairs() { return max_posy/2+1; }
int get_hi_gen() { return hi_gen; }
int get_num_gens() { return hi_gen+1; }
//table_item **operator[](int i) { return item[i]; }     //access to print, etc.
sort_table_from(tab_person *anc,int extending=0);
void print(char *s,int r,int c,int l);
ntlines(int r,int c);
nglines(int r,int c);
void clear();
table(int h=MAXGEN,int w=MAXWID);
~table();
get_couple_seed(tab_person *par1,tab_person *par2,int miny,which_marriage marriage,TABLE_ROW hi_row);
get_seed(tab_person *anc,int miny=0,int following=0);
int do_child(tab_person *kid,int miny,int pargennum,int& chminy,int& chmaxy);
};

class tab_pedigree : public pedigree
{
public:
void clear_draw_data(); /* just clears drawing data */
add_member(tab_person*temp) { return pedigree::add_member((person*)temp); }
save(data_file& file);
data_object *new_own_type();
tab_person *member(int i) { return (tab_person*) pedigree::member(i); }
int edit(int i,editable_tab_person& p) { return member(i)->revise(p); }
int read_ged_fams(data_file& file);
};

class pedparms : public data_object
{
public:
short border_w,border_h;
double ps_x_scale,ps_y_scale,ps_x_trans,ps_y_trans;
char ps_landscape;
char extend,no_diag,unsex,inset;
char default_output_type;
short gen_to_print,max_gen,max_wid;
dcstring save_file;
dcstring input_file_name;
short sym_h,sym_w,gap_h,gap_w,extra_line_h;
char title_str[65];
short t_lines[ROWS+1];
char t_line_str[41];
load(char *fn,char *ver);
save(char *fn,char *ver);
pedparms();
~pedparms();
pedparms& operator=(const pedparms& old);
int revise(editable_pedparms &);
};

extern FILE *input_file;

#ifndef wx_msw
#define LEGAL_OUTPUT_TYPES "PXS"
#else
#define LEGAL_OUTPUT_TYPES "PBIS"
#endif

#endif

