/* DCH.CXX */
/* Copyright Dave Curtis 1994 */
/* dcurtis@hgmp.mrc.ac.uk */
/* no warranty or liability of any kind is accepted, expressed or implied */

#include "dcerror.hpp"
#include "dch.hpp"
#include "winmenu.hpp"
#include "dbstate.hpp"
#include "dcutils.hpp"
#include <ctype.h>

#ifndef using_wx
dcwindow_set dch_ws(1);
#endif

#define MAX_FIELDS_FOR_IMPORT 1000
#define MAX_BUFF_LEN 5000

#ifndef min
#define min(x,y) ((x)<(y)?(x):(y))
#endif

int dch::recalc_all(char *cond)
{
if (cond!=NULL && cond[0]=='\0') cond=NULL;
db_express expr(buffer);
if (cond && !expr.parse(cond)) return 0;
if (!goto_first()) return 0;
do {
   if (!header->record_load(*buffer)) return 0;
   if (cond)
     {
     dcexpr_val *rv=expr.eval();
     float OK=(double)(*rv);
     delete rv;
     if (!OK) continue;
     }
   if (!recalculate(EVAL_ON_MODIFY)) return 0;
   if (!header->record_save(*buffer)) return 0;
   } while (goto_next() && !dcerror.stat());
return 1;
}

int dch::empty()
{
db3_file.close();
header->num_recs=0;
if (!db3_file.open(db3_fn,"w+b")) return 0;
return header->save();
}

int dch::export_dbf(char *fn,char *cond)
{
if (cond!=NULL && cond[0]=='\0') cond=NULL;
data_file f,*real_file;
db_express expr(buffer);
if (cond && !expr.parse(cond)) return 0;
if (!f.open(fn,"wb")) return 0;
real_file=header->file;
header->file=&f;
header->save();
header->save_num_recs(0L);
header->file=real_file;
f.close();
if (!f.open(fn,"r+b")) return 0;
db3_header *new_header=new db3_header(&f);
if (!new_header) return dcerror(ENOMEM);
if (!new_header->load()) return 0;
long n=0L;
int local_OK=1;
if (goto_first())
  do {
  if (!header->record_load(*buffer)) { local_OK=0; break; }
  if (cond)
    {
    dcexpr_val *rv=expr.eval();
    float use=(double)(*rv);
    delete rv;
    if (!use) continue;
    }
  if (!new_header->record_save(*buffer,++n)) { local_OK=0; break; }
  } while (goto_next());
else local_OK=0;
new_header->save_num_recs(n);
delete new_header;
return local_OK;
}

int dch::export_sdf(char *fn,char *cond)
{
if (cond!=NULL && cond[0]=='\0') cond=NULL;
data_file f;
db_express expr(buffer);
if (cond && !expr.parse(cond)) return 0;
if (!f.open(fn,"w")) return 0;
if (goto_first())
  do {
  if (!header->record_load(*buffer)) return 0;
  if (cond)
    {
    dcexpr_val *rv=expr.eval();
    float use=(double)(*rv);
    delete rv;
    if (!use) continue;
    }
  if (fwrite(buffer->buff,header->data_len-1,1,header->get_fp(f))!=1
       || f.fprintf("\n")<1)
    return dcerror(1,"Problem writing file %s",fn);
  } while (goto_next());
else return 0;
return 1;
}

int dch::import_dbf_matching_fields(char *fn,int overwrite,char *cond)
{
data_file f;
int i,table[MAX_FIELDS_FOR_IMPORT];
if (!f.open(fn,"rb")) return 0;
db3_header new_h(&f);
// !!! handle memo files later !!!
if (!new_h.load()) return 0;
db3_record_buffer new_b(&new_h);
db_express expr(&new_b);
if (cond!=NULL && cond[0]=='\0') cond=NULL;
if (cond && !expr.parse(cond)) return 0;
for (i=0;i<header->num_fields;++i)
  table[i]=new_h.flist.which(header->flist[i]->title);
if (overwrite) empty();
for (long r=1L;r<=new_h.num_recs;++r)
  {
  if (new_h.rec_is_deleted(r)) continue;
  if (!new_h.record_load(new_b,r)) return 0;
  if (cond)
    {
    dcexpr_val *rv=expr.eval();
    float OK=(double)(*rv);
    delete rv;
    if (!OK) continue;
    }
  buffer->clear();
  for (i=0;i<header->num_fields;++i)
     if (table[i]!=-1)
       memmove(buffer->operator[](i).data,new_b[table[i]].data,
            min(buffer->operator[](i).len,new_b[table[i]].len));
  header->record_save(*buffer,++header->num_recs);
  }
header->save_num_recs(header->num_recs);
return 1;
}

int dch::import_dbf(char *fn,int overwrite,char *cond)
{
data_file f;
if (!f.open(fn,"rb")) return 0;
db3_header new_h(&f);
// !!! handle memo files later !!!
if (!new_h.load()) return 0;
db3_record_buffer new_b(&new_h);
db_express expr(&new_b);
if (cond!=NULL && cond[0]=='\0') cond=NULL;
if (cond && !expr.parse(cond)) return 0;
if (overwrite) empty();
for (long r=1L;r<=new_h.num_recs;++r)
  {
  if (new_h.rec_is_deleted(r)) continue;
  if (!new_h.record_load(new_b,r)) return 0;
  if (cond)
    {
    dcexpr_val *rv=expr.eval();
    float OK=(double)(*rv);
    delete rv;
    if (!OK) continue;
    }
  buffer->clear();
  memmove(buffer->buff,new_b.buff,min(header->data_len-1,new_h.data_len-1));
  header->record_save(*buffer,++header->num_recs);
  }
header->save_num_recs(header->num_recs);
return 1;
}

#if 0
dch::import_dbf(char *fn,int overwrite,char *cond)
{
data_file f;
int i;
db3_header *new_h;
db3_record_buffer *new_b;
db_express *expr;
if (!f.open(fn,"rb")) return 0;
new_h=new db3_header(&f);
// !!! handle memo files later !!!
if (!new_h->load()) { delete new_h; return 0; }
new_b=new db3_record_buffer(new_h);
if (!new_b) { delete new_h; return dcerror(ENOMEM); }
expr=new db_express(new_b);
if (!expr) { delete new_b; delete new_h; return dcerror(ENOMEM); }
if (cond!=NULL && cond[0]=='\0') cond=NULL;
if (cond && !expr->parse(cond))
  { delete expr; delete new_b; delete new_h; return 0; }
if (overwrite) empty();
for (long r=1L;r<=new_h->num_recs;++r)
  {
  if (new_h->rec_is_deleted(r)) continue;
  if (!new_h->record_load(*new_b,r))
      { delete expr; delete new_b; delete new_h; return 0; }
  if (cond)
    {
    dcexpr_val *rv=expr->eval();
    float OK=(double)(*rv);
    delete rv;
    if (!OK) continue;
    }
  buffer->clear();
  memmove(buffer->buff,new_b->buff,min(header->data_len-1,new_h->data_len-1));
  header->record_save(*buffer,++header->num_recs);
  }
header->save_num_recs(header->num_recs);
delete expr; delete new_b; delete new_h;
return 1;
}
#endif

int dch::import_sdf(char *fn,int overwrite)
{
data_file f;
char line_buff[MAX_BUFF_LEN];
if (!f.open(fn,"r")) return 0;
if (overwrite) empty();
while (f.fgets(line_buff,MAX_BUFF_LEN))
  {
  char *ptr;
  if ((ptr=strchr(line_buff,'\n'))!=NULL) *ptr='\0';
  buffer->clear();
  strncpy(buffer->buff,line_buff,min(strlen(line_buff),header->data_len-1));
  header->record_save(*buffer,++header->num_recs);
  }
header->save_num_recs(header->num_recs);
return 1;
}

int dch::import_sdf_by_record(char *fn)
{
data_file f;
char line_buff[MAX_BUFF_LEN],rest[MAX_BUFF_LEN],idx_fn[100],field_name[11];
int i,field_num[MAX_FIELDS_FOR_IMPORT],pos[MAX_FIELDS_FOR_IMPORT],len[MAX_FIELDS_FOR_IMPORT],nfields;
if (!f.open(fn,"r")) return 0;
if (!f.fgets(line_buff,MAX_BUFF_LEN))
  return dcerror(1,"%s is empty!",fn);
*rest='\0';
if (sscanf(line_buff,"%s %d %d %[^\n]",idx_fn,&pos[0],&len[0],rest)<3)
  return dcerror(1,"%s has bad first line:\n %s",fn,line_buff);
nfields=1;
while (strcpy(line_buff,rest),*rest='\0',
  sscanf(line_buff,"%s %d %d %[^\n]",field_name,&pos[nfields],&len[nfields],rest)>=3)
  {
  if ((field_num[nfields]=header->flist.which(field_name))==-1)
    return dcerror(1,"Unrecognised field name %s in %s",field_name,fn);
  ++nfields;
  }
dc_index idx;
if (!idx.open_old(idx_fn))
  return dcerror(1,"Could not open index file %s",idx_fn);
while (f.fgets(line_buff,MAX_BUFF_LEN))
  {
  long r;
  int line_len=strlen(line_buff);
  if (line_len<pos[0]-1+len[0]) continue;
  strcpy(rest,line_buff+pos[0]-1);
  rest[len[0]]='\0';
  if ((r=idx.exact_find(rest))==0L)
//    return dcerror(1,"No match for key: %s, in line: %s",rest,line_buff);
    return dcerror(1,"No match for key: %s",rest);
  else if (!header->goto_rec(r) || !header->record_load(*buffer))
    return dcerror(1,"Problem reading record");
  for (i=1;i<nfields;++i)
    {
    if (line_len<pos[i]-1) break;
    strcpy(rest,line_buff+pos[i]-1);
    rest[len[i]]='\0';
    buffer->add_string(field_num[i],rest);
    }
  if (!header->record_save(*buffer))
    return dcerror(1,"Problem saving record");
  }
return 1;
}

int dch::run_report(char *rep_fn,char *rep_out_fn,char *crit)
{
db_report rep;
data_file rep_file,out_file;
if (!rep_file.open(rep_fn,"r") || !out_file.open(rep_out_fn,"w")) return 0;
db_report_outfile out(&out_file);
return (rep.set_up(&out,buffer,&rep_file) && rep.run(get_curr_index(),crit));
}

int dch::choose_index(winmenu &m)
{
int i,rv=0;
dczGDList(char) names(0);
index_list.start();
for (i=0;i<(int)index_list.size();++i,++index_list)
  names<<(char*)index_list()->filename;
names<<"NO INDEX";
int c;
while ((c=m.choose(names))!=0)
  {
  if (c==(int)index_list.size()+1)
    { curr_index=NULL; rv=1; break; }
  index_list.start();
  for (i=1;i<c;++i) ++index_list;
  if (!index_list()->is_OK()
    && index_list()->idx.open_old((char*)index_list()->filename)) // failure will generate an error message
      index_list()->parse(buffer);
  if (index_list()->is_OK())
    { curr_index=index_list(); rv=1; break; }
  }
m.suspend();
return rv;
}

int dch::search_for_expr(char *s)
{
if (s)
  {
  if (search_expr) delete search_expr;
  if ((search_expr=new db_express(buffer))==NULL) return dcerror(ENOMEM);
  if (!search_expr->parse(s)) return 0;
  }
else if (!search_expr)
  return dcerror(1,"No search expression has been specified");
double match=0.0;
for (;;)
  {
  if ((s==NULL && !goto_next()) || (s!=NULL && !goto_first())
     || !header->record_load(*buffer))
    {
    if (s==NULL) goto_first(); // position back at start after failure
    if (search_expr) delete search_expr;
    search_expr=NULL;
    return 0; // finished going through all records
    }
  s=NULL; // no longer first time
  dcexpr_val *rv;
  rv=search_expr->eval();
  match=double(*rv);
  delete rv;
  if (match) return 1;
  }
}

int dch::recalculate(int when)
{
long i;
editor->fieldlist.start();
for (i=0L;i<editor->fieldlist.size();++i,++editor->fieldlist)
    {
    if (editor->fieldlist()->get_len()==0) continue; // allow prompts alone
    db3_field *f=(db3_field *)editor->fieldlist(); // it better be one
    if (f->when_to_eval & when)
      {
      dcexpr_val *v=f->eval(buffer);
      if (v==NULL) return 0;  // horrible memory problems
      char *s=(char *)(*v);
      buffer->add_string(f->field_num,s);
      delete v;
      }
    }
return 1;
}

int dch::edit_one(int allow_next_or_prev)
{
// data has already been read in to buffer
dcexpr_val *new_str,*old_str[MAXINDICES];
long which_rec=header->get_this_rec();
char *save_buff;
if (which_rec<1L) return 0;
if ((save_buff=new char[header->data_len-1])==0) return dcerror(ENOMEM);
strncpy(save_buff,buffer->buff,header->data_len-1);
int i,rv=0;
index_list.start();
for (i=0L;i<(int)index_list.size();++i,++index_list)
  {
  if (!index_list()->is_OK()) { old_str[i]=0; continue; }
  else old_str[i]=index_list()->expr->eval();
  }
if ((rv=editor->do_editing(buffer,allow_next_or_prev))!=0
    && strncmp(save_buff,buffer->buff,header->data_len-1))
  {
  recalculate(EVAL_ON_MODIFY);
  header->record_save(*buffer,which_rec);
  index_list.start();
  for (i=0L;i<(int)index_list.size();++i,++index_list)
    {
    if (!index_list()->is_OK()) continue;
    new_str=index_list()->expr->eval();
    if (strcmp((char*)*old_str[i],(char*)*new_str))
      {
      if (index_list()->idx.find_matching_node((char*)*old_str[i],which_rec)) // ought to
        index_list()->idx.remove();
      index_list()->idx.add((char*)*new_str,which_rec);
      if (index_list()==curr_index)
        index_list()->idx.find_matching_node((char*)*new_str,which_rec);
        // make record current in new position
      }
    delete new_str;
    }
  }
for (i=0L;i<(int)index_list.size();++i)
  if (old_str[i]) delete old_str[i];
delete save_buff;
return rv;
}

int dch::delete_one(int allow_next_or_prev)
{
// dcexpr_val *new_str,*old_str[MAXINDICES];
int i,rv=0;
long which_rec=header->get_this_rec();
if (which_rec<1L) return 0;
if ((rv=editor->do_editing(buffer,allow_next_or_prev))!=0)
  {
  buffer->del();
  header->record_save(*buffer);
  index_list.start();
  for (i=0L;i<(int)index_list.size();++i,++index_list)
    if (index_list()->is_OK())
      {
      dcexpr_val *new_str=index_list()->expr->eval();
      if (index_list()->idx.find_matching_node((char*)*new_str,which_rec)) // ought to
      // ought to
        index_list()->idx.remove();
      delete new_str;
      }
  if (rv==2 && !goto_prev()) { goto_next(); rv=1; }
  if ((rv==1 || rv==3) && !goto_next()) { goto_prev(); rv=1; }
  }
return rv;
}

int dch::add_one(int allow_next_or_prev)
{
dcexpr_val *new_str;
int i,rv=0;
buffer->clear();
if ((rv=editor->do_editing(buffer,allow_next_or_prev))!=0)
  {
  header->goto_rec(header->num_recs+1L); // maybe in case can use record number in calcs
  recalculate(EVAL_ON_ADD);
  header->record_save(*buffer,header->num_recs+1L);
  ++header->num_recs;
  header->save_num_recs(header->num_recs);
  index_list.start();
  for (i=0L;i<(int)index_list.size();++i,++index_list)
    {
    if (!index_list()->is_OK()) continue;
    new_str=index_list()->expr->eval();
    index_list()->idx.add((char*)*new_str,header->num_recs);
    if (index_list()==curr_index)
        index_list()->idx.find_matching_node((char*)*new_str,header->num_recs);
    delete new_str;
    }
  }
return rv;
}

int dch::undelete_one(int allow_next_or_prev)
{
dcexpr_val *new_str;
int i,rv=0;
header->record_load(*buffer);
if ((rv=editor->do_editing(buffer,allow_next_or_prev))!=0)
  {
  buffer->un_del();
  header->record_save(*buffer);
  index_list.start();
  for (i=0L;i<(int)index_list.size();++i,++index_list)
    {
    if (!index_list()->is_OK()) continue;
    new_str=index_list()->expr->eval();
    index_list()->idx.add((char*)*new_str,header->this_rec);
    if (index_list()==curr_index)
        index_list()->idx.find_matching_node((char*)*new_str,header->this_rec);
    delete new_str;
    }
  }
return rv;
}

dc_index_file::dc_index_file(char *f,char *e) : filename(f),expr_str(e)
{ expr=NULL; }

dc_index_file::~dc_index_file()
{ idx.close(); if (expr) delete expr; }

char *GETPARAM(char *s,char *p)
{ return strncmp(s,p,strlen(p))?NULL:s+strlen(p); }

int dch::dbf_from_hdc(char *fn)
{
data_file f,temp_db3_file;
dcstring temp_db3_fn;
char fn_buff[200],*line_buff,real_line_buff[2000],*ptr;
if (!f.open(fn,"r")) return 0;
strcpy(real_line_buff,fn);
if ((ptr=strchr(real_line_buff,'.'))==NULL) ptr=strchr(real_line_buff,'\0');
strcpy(ptr,".dbf");
if (!temp_db3_fn.assign(real_line_buff)) return dcerror(ENOMEM);
// default DBF file name is same as HDC file
for (;;)
  {
  if (!f.fgets(real_line_buff,2000))
    return dcerror(1,"No fields defined in HDC file %s",fn);
  else
    {
    line_buff=real_line_buff;
    while (*line_buff&&isspace(*line_buff)) ++line_buff;
    }
  if ((ptr=GETPARAM(line_buff,"DBF="))!=NULL)
    {
    if (sscanf(ptr,"%s",fn_buff)!=1) return dcerror(1,"No DBF filename: %s",line_buff);
    else if (!temp_db3_fn.assign(fn_buff)) return dcerror(ENOMEM);
    }
  else if ((ptr=GETPARAM(line_buff,"INDEX="))!=NULL)
    {
    char expr_buff[2000];
    if (sscanf(ptr," %s %[^\n]",fn_buff,expr_buff)<2)
       return dcerror(1,"Could not read legal index file definition: ",ptr);
    dc_index_file ifile(fn_buff,expr_buff);
    ifile.idx.make_new(fn_buff);
    ifile.idx.close();
    }
  else if (*line_buff=='N') break;
  }
FILE *test;
if ((test=fopen(temp_db3_fn,"rb"))!=NULL)
  {
  fclose(test);
  return dcerror(1,"File %s already exists",(char*) temp_db3_fn);
  }
if (!temp_db3_file.open(temp_db3_fn,"w+b")) return 0;
db3_header *temp_header=new db3_header(&temp_db3_file);
if (!temp_header) return dcerror(ENOMEM);
char first=1,type='C',len=0,done=0;
// char *target;
dcstring field_name("");
while (!done)
  {
  char temp_buff[1000];
  switch (*line_buff)
    {
    case 'N':
    case '\0':
      if (first)
        {
        first=0;
        *temp_buff='\0';
        sscanf(line_buff+1,"%s",temp_buff);
        field_name.assign(temp_buff);
        break;
        }
      if (*(char*)field_name=='\0') break;  // just prompt
      temp_header->add_field(field_name,type,len);
#if 0
      if (type=='M')
        {
        db3_memo_field *mf=
         new db3_memo_field(row,col,prompt,target,view,buffer->header->memofile);
        fieldlist<<mf;
        memo_list<<mf;
        ++nmemos;
        }
      else
#endif
      type='C';
      if (*line_buff=='\0') done=1;
      else
        {
        *temp_buff='\0';
        sscanf(line_buff+1,"%s",temp_buff);
        field_name.assign(temp_buff);
        }
      break;
    case 'T':
      type=line_buff[1];
      break;
    case 'L':
      len=atoi(line_buff+1);
      break;
    }
  line_buff=real_line_buff;
  if (!f.fgets(real_line_buff,2000) || *real_line_buff=='|' || *real_line_buff=='Z')
     *real_line_buff='\0';
  else
    {
    while (*line_buff && isspace(*line_buff)) ++line_buff;
    if (!*line_buff) continue; // blank line
    }
  }
return temp_header->save();
}

#ifndef using_wx
int dch::load(char *fn,dcwindow_set *ws)
#else
int dch::load(char *fn,wxFrame **fptr)
#endif
{
OK=0;
data_file f;
char fn_buff[200],*line_buff,real_line_buff[2000],*ptr;
extern dcstring lu_hdr_fn,lu_idc_fn,lu_key;
lu_hdr_fn.assign(""),lu_idc_fn.assign(""),lu_key.assign("");
// clear these now or live to regret it
if (!f.open(fn,"r")) return 0;
strcpy(real_line_buff,fn);
if ((ptr=strrchr(real_line_buff,'.'))==NULL) ptr=strchr(real_line_buff,'\0');
strcpy(ptr,".dbf");
if (!db3_fn.assign(real_line_buff)) return dcerror(ENOMEM);
while (index_list.size()) delete *index_list;
curr_index=NULL;
for (;;)
  {
  if (!f.fgets(real_line_buff,2000))
    return dcerror(1,"No fields defined in HDC file %s",fn);
  else
    {
    line_buff=real_line_buff;
    while (isspace(*line_buff)) ++line_buff;
    }
  if ((ptr=GETPARAM(line_buff,"DBF="))!=NULL)
    {
    if (sscanf(ptr,"%s",fn_buff)!=1) return dcerror(1,"No DBF filename: %s",line_buff);
    else if (!db3_fn.assign(fn_buff)) return dcerror(ENOMEM);
    }
  else if ((ptr=GETPARAM(line_buff,"TOP="))!=NULL) top=atoi(ptr);
  else if ((ptr=GETPARAM(line_buff,"LEFT="))!=NULL) left=atoi(ptr);
  else if ((ptr=GETPARAM(line_buff,"WIDE="))!=NULL) wide=atoi(ptr);
  else if ((ptr=GETPARAM(line_buff,"HIGH="))!=NULL) high=atoi(ptr);
  else if ((ptr=GETPARAM(line_buff,"INDEX="))!=NULL)
    {
    char expr_buff[2000];
    if (sscanf(ptr," %s %[^\n]",fn_buff,expr_buff)<2)
       return dcerror(1,"Could not read legal index file definition: ",ptr);
    dc_index_file *ifile=new dc_index_file(fn_buff,expr_buff);
    if (!ifile) return dcerror(ENOMEM);
    index_list<<ifile;
    }
  else if (*line_buff=='N') break;
  }
if (!db3_file.open(db3_fn,"r+b")) return 0;
if (header) delete header;
header=new db3_header(&db3_file);
if (!header) return dcerror(ENOMEM);
if (!header->load()) return 0;
if (buffer) delete buffer;
buffer=new db3_record_buffer(header);
if (!buffer) return dcerror(ENOMEM);
if (editor) delete editor;
#ifndef using_wx
if ((editor=new db3_editor(header,ws?ws:&dch_ws,db3_fn,top,left,top+high+1,left+wide+1))==NULL)
#else
if ((editor=new db3_editor(header,fptr,db3_fn,top,left,top+high+2,left+wide+1))==NULL)
#endif
  return dcerror(ENOMEM);
char first=1,type='C',done=0;
char *target;
int len=-1,row,col,view=-1,maxrow=0,field_num;
dcstring prompt(""),field_name(""),expr_str("");
while (!done)
  {
  char temp_buff[1000];
  switch (*line_buff)
    {
    case '(':
      expr_str.assign(line_buff);
      break;
    case 'N':
    case '\0':
      if (first)
        {
        first=0;
        *temp_buff='\0';
        sscanf(line_buff+1,"%s",temp_buff);
        field_name.assign(temp_buff);
        prompt.assign(temp_buff);
        break;
        }
      if (*(char*)field_name=='\0')
        {
        if (*(char*)prompt=='\0')
          return dcerror(1,"Field with no name or prompt before line %s",
            line_buff);
        else editor->fieldlist <<
          new edit_field(row,col,prompt,NULL,0);
        }
      else
        {
        field_num=buffer->header->flist.which(field_name);
        if (field_num==-1)
          return dcerror(1,"Field %s is not in file %s",(char*)field_name,(char*)db3_fn);
        db3_field_defn* def=buffer->header->flist[field_num];
        target=editor->buff+def->offset;
        if (len==-1) len=def->length;
        else if (len!=def->length)
          return dcerror(1,"Length incorrect for field %s (should be %d)",
            (char*)field_name,(int)def->length);
        if (view==-1) view=len;
        if (view>wide-col-3) view=wide-col-3;
#if 0
      if (type=='M')
        {
        db3_memo_field *mf=
         new db3_memo_field(row,col,prompt,target,view,buffer->header->memofile);
        fieldlist<<mf;
        memo_list<<mf;
        ++nmemos;
        }
      else
#endif
      if (type=='D')
          editor->fieldlist <<
            new db3_date_field(row,col-strlen(prompt),prompt,target,view,
               len,(char*)expr_str,field_num);
      else
          editor->fieldlist <<
            new db3_edit_field(row,col-strlen(prompt),prompt,target,view,
               len,(char*)expr_str,field_num);
        }
      expr_str.assign("");
      prompt.set_null(); // stops prompt string being deleted, left for
      // ever in edit_field and never deleted - to fix one day
      field_num=-1;
      len=view=-1;
      type='C';
      if (*line_buff=='\0') done=1;
      else
        {
        sscanf(line_buff+1,"%s",temp_buff);
        field_name.assign(temp_buff);
        prompt.assign(temp_buff);
        }
      break;
    case 'P':
      *temp_buff='\0';
      sscanf(line_buff+1,"%[^\n]",temp_buff);
      prompt.assign(temp_buff);
      break;
    case 'T':
      type=line_buff[1];
      break;
    case 'L':
      len=atoi(line_buff+1);
      break;
    case 'C':
      col=atoi(line_buff+1)-1;
      if (col<0)
          return dcerror(1,"Minimum value for column position is 1");
      break;
    case 'R':
      row=atoi(line_buff+1)-1;
      if (row>maxrow) maxrow=row;
      if (row<0)
          return dcerror(1,"Minimum value for row position is 1");
      break;
    case 'V':
      view=atoi(line_buff+1);
      break;
    }
  if (!f.fgets(real_line_buff,2000) || *line_buff=='|' || *line_buff=='Z')
     *line_buff='\0';
  else
    {
    line_buff=real_line_buff;
    while (isspace(*line_buff)) ++line_buff;
    }
  }
editor->set_npages(maxrow/high+1);
OK=1;
index_list.start();
for (long i=0;i<index_list.size();++i,++index_list)
    {
    if (index_list()->idx.open_old(index_list()->filename) // allowed to fail
        && index_list()->parse(buffer) && curr_index==NULL)
           curr_index=index_list();
// but main point of this code is to open and parse indices
    }
buffer->clear();
return 1;
}

dch::dch()
{
top=0; left=0; wide=78;
#ifndef using_wx
high=21;
#else
high=18;
#endif
editor=NULL;
buffer=NULL;
header=NULL;
curr_index=NULL;
search_expr=NULL;
OK=0;
}

dch::~dch()
{
if (editor) delete editor;
if (buffer) delete buffer;
if (header) delete header;
while (index_list.size()) delete *index_list;
}

int dc_index_file::parse(db3_record_buffer *b)
{
if (expr!=NULL) delete expr;
if ((expr=new db_express(b))==NULL) return dcerror(ENOMEM);
if (!expr->parse((char*)expr_str))
  { delete expr; expr=NULL; return 0; }
else return 1;
}

int dch::rebuild_indices()
{
long i;
for (i=0L,index_list.start();i<index_list.size();++i,++index_list)
  {
  if (!index_list()->idx.make_new(index_list()->filename)) return 0;
  if (!index_list()->parse(buffer)) return 0;
  }
for (long r=1L;r<=buffer->header->get_num_recs();++r)
  {
  if (header->rec_is_deleted(r)) continue;
  header->record_load(*buffer,r);
  for (i=0L,index_list.start();i<index_list.size();++i,++index_list)
      {
      dcexpr_val *rv=index_list()->expr->eval();
      if (dcerror.stat() || !index_list()->idx.add((char*)(*rv),r))
        {
        delete rv;
        index_list()->idx.close();
        return 0;
        }
      delete rv;
      }
  }
if (curr_index==NULL)  //set to default index
  {
  index_list.start();
  curr_index=index_list();
  }
return 1;
}

int db3_editor::check()
{
return 0;
}

int db3_editor::copy_in(void *source)
{
db3_record_buffer *s=(db3_record_buffer *)source;
memcpy(buff,s->buff,header->data_len-1);
return 1;
}

int db3_editor::copy_out(void *dest)
{
db3_record_buffer *d=(db3_record_buffer *)dest;
memcpy(d->buff,buff,header->data_len-1);
return 1;
}

#ifndef wx_x
db3_editor::db3_editor(db3_header *h,dcwindow_set *ws,char *t,
  int r1,int c1,int r2,int c2,int att,int fr,int fratt,void (*hf)(char *form,...))
  : db_editor(ws,t,r1,c1,r2,c2,1,att,fr,fratt,hf), db3_record_buffer(h)
{ ; }
#else
db3_editor::db3_editor(db3_header *h,wxFrame **fptr,char *t,
  int r1,int c1,int r2,int c2)
  : db_editor(fptr,t,r1,c1,r2,c2,1), db3_record_buffer(h)
{ ; }
#endif

dcexpr_val *db3_field::eval(db3_record_buffer *b)
{
if (expr!=NULL) return expr->eval();
else if ((expr=new db_express(b))==NULL)
   { dcerror(ENOMEM); return NULL; } // returning NULL may cause problems
else
  {
  if (!expr->parse(expr_str)) expr->parse("\"BAD EXPRESSION\"");
  return expr->eval();
  }
}

db3_field::db3_field(int r,int c, char *p,char *t,int l,int rl,char *e,int fnum) :
  edit_field(r,c,p,t,l,rl), expr_str(e)
{
field_num=fnum;
expr=NULL;
when_to_eval=0;
int i;
for (i=strlen(expr_str)-1;i>=0 && isspace(expr_str[i]);--i)
    expr_str[i]='\0'; // truncate white space
if (i>0)  // must be at least two parentheses + legal expression
    {
    if (expr_str[i-1]=='*')
      {
      if (expr_str[i]=='A') when_to_eval=EVAL_ON_ADD;
      else if (expr_str[i]=='M') when_to_eval=EVAL_ON_MODIFY;
      }
    else when_to_eval=EVAL_ON_ADD | EVAL_ON_MODIFY;
    }
}

db3_date_field::db3_date_field(int r,int c, char *p,char *t,int l,int rl,
  char *e,int fnum) : db3_edit_field(r,c,p,t,l,rl,e,fnum) {;}

db3_edit_field::db3_edit_field(int r,int c, char *p,char *t,int l,int rl,
  char *e,int fnum) : db3_field(r,c,p,t,l,rl,e,fnum) {;}

db3_field::~db3_field()
{ if (expr) delete expr; }

int db3_edit_field::is_memo()
  { return 0; }

int db3_date_field::string_to_date(char *date,char *s)
{
int d=0,y=0,m=0,c;
c=sscanf(s,"%d %*c %d %*c %d",&d,&m,&y);
if (y<100) y+=1900;
sprintf(date,"%04d%02d%02d",y,m,d);
return c==3;
}

int db3_date_field::date_to_string(char *s,char *date)
{
int d=0,y=0,m=0,c;
c=sscanf(date,"%4d%2d%2d",&y,&m,&d);
// sprintf(s,"%02d/%02d/%02d",d,m,y%100);
sprintf(s,"%02d/%02d/%04d",d,m,y);
return c==3;
}

#ifdef using_wx
void db3_edit_field::encode()
{
char *s=((wxText*)wx_editor)->GetValue();
strncpy((char*)target,s,len);
for (int i=strlen((char*)target);i<len;++i)
  *((char*)target+i)=' ';
}

void db3_edit_field::decode()
{
char buff[200];
strncpy(buff,(char*)target,len);
buff[len]='\0';
for (int i=len-1;i>=0 && buff[i]==' ';--i)
  buff[i]='\0';
((wxText*)wx_editor)->SetValue(buff);
}

void db3_date_field::encode()
{
char date_buff[100];
char *s=((wxText*)wx_editor)->GetValue();
string_to_date(date_buff,s);
strncpy((char*)target,date_buff,len);
for (int i=strlen((char*)target);i<len;++i)
  *((char*)target+i)=' ';
}

void db3_date_field::decode()
{
char buff[200],date_str[200];
strncpy(buff,(char*)target,len);
buff[len]='\0';
date_to_string(date_str,buff);
((wxText*)wx_editor)->SetValue(date_str);
}
#else
void db3_edit_field::decode()
{
strncpy(buffer,target,len); buffer[len]='\0';
for (int i=len-1;i>=0 && buffer[i]==' ';--i)
  buffer[i]='\0';
}

void db3_edit_field::encode()
{
strncpy(target,buffer,len);
for (int i=strlen((char*)target);i<len;++i)
  *((char*)target+i)=' ';
}

void db3_date_field::decode()
{
char date_buff[100];
strncpy(date_buff,target,len); date_buff[len]='\0';
date_to_string(buffer,date_buff);
}

void db3_date_field::encode()
{
char date_buff[100];
string_to_date(date_buff,buffer);
strncpy(target,date_buff,len);
for (int i=strlen((char*)target);i<len;++i)
  *((char*)target+i)=' ';
}
#endif

#if 0
db3_memo_field::db3_memo_field(int r,int c, char *p,char *t,int l,db3_memo_file *mf,
      int r1,int c1,int r2,int c2,int att,int hatt,PFC hf,
      int ml,int lm,int rm)
  : (r,c,p,t,l,l), editor(r1,c1,r2,c2,att,hatt,hf,ml,lm,rm)
{
file=mf;
}

void db3_memo_field::encode() { ; }

void db3_memo_field::decode() { ; }

int db3_memo_field::is_memo()
  { return 1; }

#endif



